from copy import deepcopy
from django.db import models
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property


class QuerysetCache(dict):
    def __init__(self, queryset):
        self._queryset = queryset

        # It's possible the queryset is prefetched with prefetch_attribute_values.
        # In this case the queryset is actually a list, and thus we can't use .iterator(),
        if isinstance(queryset, list):
            self._queryset_iterator = queryset
        else:
            self._queryset_iterator = queryset.iterator()

    def queryset(self):
        return self._queryset

    def get(self, code, default=None):
        try:
            return self[code]
        except KeyError:
            for instance in self._queryset_iterator:
                instance_code = instance.code
                self[instance_code] = instance
                if instance_code == code:
                    return instance

        return default


class AttributesQuerysetCache:
    def __init__(self, product):
        self.product = product

    @cached_property
    def attributes(self):
        return QuerysetCache(self.product.get_product_class().attributes.all())

    @cached_property
    def attribute_values(self):
        # This means this product comes from a prefetched queryset with the
        # prefetch_attribute_values method. Which selects the attribute and
        # annotates the attribute code. This avoids the need of extra queries.
        if hasattr(self.product, "_prefetched_attribute_values"):
            return QuerysetCache(self.product.get_attribute_values())

        return QuerysetCache(
            self.product.get_attribute_values()
            .select_related("attribute")
            .annotate(code=models.F("attribute__code"))
        )

    def set_attributes(self, attributes):
        self.__dict__["attributes"] = attributes


class ProductAttributesContainer:
    """
    Stolen liberally from django-eav, but simplified to be product-specific

    To set attributes on a product, use the `attr` attribute:

        product.attr.weight = 125

    To refetch the attribute values from the database:

        product.attr.refresh()
    """

    RESERVED_ATTRIBUTES = {
        "_cache",
        "_dirty",
        "initialized",
        "_initialized",
        "_product",
        "get_all_attributes",
        "get_attribute_by_code",
        "get_value_by_attribute",
        "get_values",
        "initialize",
        "product",
        "refresh",
        "prepare_save",
        "save",
        "set",
        "update",
        "validate_attributes",
    }

    def __init__(self, product):
        # use __dict__ directly to avoid triggering __setattr__, which would
        # cause a recursion error on _initialized.
        self.__dict__.update(
            {
                "_product": product,
                "_initialized": False,
                "_dirty": set(),
                "_cache": None,
            }
        )

    def __deepcopy__(self, memo):
        cpy = ProductAttributesContainer(self.product)
        memo[id(self)] = cpy
        if self.initialized:
            # Only copy attributes for initialized containers
            for key, value in self.__dict__.items():
                if key != "_cache":
                    cpy.__dict__[key] = deepcopy(value, memo)

        return cpy

    def __getstate__(self):
        # Allow everything to go into the pickle except for _cache (which can't
        # be pickled since it contains a generator)
        d = {}
        d.update(self.__dict__)
        d["_cache"] = None
        return d

    def __setstate__(self, d):
        # Update __dict__ instead of setting it to avoid triggering __setattr__,
        # which causes a recursion error.
        self.__dict__.update(d)

    @property
    def product(self):
        return self._product

    @property
    def initialized(self):
        return self._initialized

    @property
    def cache(self):
        if self.__dict__["_cache"] is None:
            self.__dict__["_cache"] = AttributesQuerysetCache(self.product)
        return self.__dict__["_cache"]

    def initialize(self):
        self.__dict__["_initialized"] = True
        # initialize should not overwrite any values that have allready been set
        attrs = self.__dict__
        for v in self.get_values():
            attrs.setdefault(v.attribute.code, v.value)

    def invalidate(self):
        "Invalidate any stored data, queried from the database"
        self.__dict__["_cache"] = None
        self.__dict__["_initialized"] = False

    def refresh(self):
        "Reload any queried data from the database, discarding local changes"
        self.__dict__["_cache"] = None
        for v in self.get_values():
            setattr(self, v.attribute.code, v.value)

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if self.initialized:
                raise
            else:
                self.initialize()
                # try again, whever happens then will just be the result.
                return super().__getattribute__(name)

    def __getattr__(self, name):
        raise AttributeError(
            _("%(obj)s has no attribute named '%(attr)s'")
            % {"obj": self.product.get_product_class(), "attr": name}
        )

    def __setattr__(self, name, value):
        if name in self.RESERVED_ATTRIBUTES:
            raise ValidationError(
                "%s is a reserved name and cannot be used as an attribute" % name
            )

        self._dirty.add(name)
        super().__setattr__(name, value)

    def set(self, name, value, validate_identifier=True):
        if not validate_identifier or name.isidentifier():
            self.__setattr__(name, value)
        else:
            raise ValidationError(
                _(
                    "%s is not a valid identifier, but attribute codes must be valid python identifiers"
                    % name
                )
            )

    def update(self, adict):
        self._dirty.update(adict.keys())
        self.__dict__.update(adict)

    def validate_attributes(self):
        for attribute in self.get_all_attributes():
            value = getattr(self, attribute.code, None)
            if value is None:
                if attribute.required:
                    raise ValidationError(
                        _("%(attr)s attribute cannot be blank")
                        % {"attr": attribute.code}
                    )
            else:
                try:
                    attribute.validate_value(value)
                except ValidationError as e:
                    raise ValidationError(
                        _("%(attr)s attribute %(err)s")
                        % {"attr": attribute.code, "err": e}
                    )

    def get_values(self):
        return self.cache.attribute_values.queryset()

    def get_value_by_attribute(self, attribute):
        return self.cache.attribute_values.get(attribute.code)

    def get_all_attributes(self):
        return self.cache.attributes.queryset()

    def get_attribute_by_code(self, code):
        return self.cache.attributes.get(code)

    def __iter__(self):
        return iter(self.get_values())

    def prepare_save(self):
        to_be_updated = []
        to_be_created = []
        to_be_deleted = []
        update_fields = set()

        if not self.initialized and not self._dirty:
            # no need to save untouched attr lists
            return (to_be_updated, to_be_created, to_be_deleted, update_fields)

        ProductAttributeValue = self.product.attribute_values.model

        for attribute in self.get_all_attributes():
            if hasattr(self, attribute.code):
                value = getattr(self, attribute.code)
                if attribute.code not in self._dirty:
                    # Make sure that if a value comes from a parent product, it is not
                    # copied to the child, we do this by checking if a value has been
                    # changed, which would not be the case if the value comes from the
                    # parent.
                    # for attributes are are set explicitly (_dirty), this check is not
                    # needed and should always be saved.
                    try:
                        attribute_value_current = self.get_value_by_attribute(attribute)
                        if (
                            attribute_value_current is not None
                            and attribute_value_current.value == value
                        ):
                            continue  # no new value needs to be saved
                    except ObjectDoesNotExist:
                        pass  # there is no existing value, so a value needs to be saved.

                if attribute.is_multi_option:  # multi_option can not be bulk saved
                    attribute.save_value(self.product, value)
                else:
                    value_obj = self.get_value_by_attribute(attribute)
                    if (
                        value_obj is None or value_obj.product != self.product
                    ):  # it doesn't exist yet so should be created
                        new_value_obj = ProductAttributeValue(
                            attribute=attribute, product=self.product
                        )

                        bound_value_obj = attribute.bind_value(new_value_obj, value)
                        # don't create attributevalues that wheren't even set at all.
                        if bound_value_obj is not None and bound_value_obj.is_dirty:
                            assert not bound_value_obj.pk
                            to_be_created.append(bound_value_obj)
                    else:
                        bound_value_obj = attribute.bind_value(value_obj, value)
                        if bound_value_obj is None:
                            to_be_deleted.append(value_obj.pk)
                        else:
                            if bound_value_obj.attribute.is_entity:
                                # entities can be bulk_created, but not bulk_saved
                                bound_value_obj.save()
                                continue

                            if bound_value_obj.attribute.is_file:
                                # with bulk_create the file is save just fine, but
                                # with buld_update, it's not, so we have to performa
                                # that manually
                                bound_value_obj._meta.get_field(
                                    bound_value_obj.value_field_name
                                ).pre_save(bound_value_obj, False)

                            if bound_value_obj.attribute_id != attribute.id:
                                bound_value_obj.attribute = attribute
                                update_fields.add("attribute")

                            to_be_updated.append(bound_value_obj)
                            update_fields.add(bound_value_obj.value_field_name)

        return (to_be_updated, to_be_created, to_be_deleted, update_fields)

    def save(self):
        to_be_updated, to_be_created, to_be_deleted, update_fields = self.prepare_save()

        # now save all the attributes in bulk
        if to_be_deleted:
            self.product.attribute_values.filter(pk__in=to_be_deleted).delete()
        if to_be_updated:
            self.product.attribute_values.bulk_update(
                to_be_updated, update_fields, batch_size=500
            )
        if to_be_created:
            self.product.attribute_values.bulk_create(
                to_be_created, batch_size=500, ignore_conflicts=False
            )

        # after this the current data is nolonger valid and should be refetched
        # from the database
        self.invalidate()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0003_auto_20150113_1629'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='line',
            options={'ordering': ['pk'], 'verbose_name': 'Order Line', 'verbose_name_plural': 'Order Lines'},
        ),
    ]



# pylint: disable=wildcard-import, unused-wildcard-import

"""
Vanilla product models
"""
from oscar.apps.catalogue.abstract_models import *
from oscar.core.loading import is_model_registered

__all__ = ["ProductAttributesContainer"]


if not is_model_registered("catalogue", "ProductClass"):

    class ProductClass(AbstractProductClass):
        pass

    __all__.append("ProductClass")


if not is_model_registered("catalogue", "Category"):

    class Category(AbstractCategory):
        pass

    __all__.append("Category")


if not is_model_registered("catalogue", "ProductCategory"):

    class ProductCategory(AbstractProductCategory):
        pass

    __all__.append("ProductCategory")


if not is_model_registered("catalogue", "Product"):

    class Product(AbstractProduct):
        pass

    __all__.append("Product")


if not is_model_registered("catalogue", "ProductRecommendation"):

    class ProductRecommendation(AbstractProductRecommendation):
        pass

    __all__.append("ProductRecommendation")


if not is_model_registered("catalogue", "ProductAttribute"):

    class ProductAttribute(AbstractProductAttribute):
        pass

    __all__.append("ProductAttribute")


if not is_model_registered("catalogue", "ProductAttributeValue"):

    class ProductAttributeValue(AbstractProductAttributeValue):
        pass

    __all__.append("ProductAttributeValue")


if not is_model_registered("catalogue", "AttributeOptionGroup"):

    class AttributeOptionGroup(AbstractAttributeOptionGroup):
        pass

    __all__.append("AttributeOptionGroup")


if not is_model_registered("catalogue", "AttributeOption"):

    class AttributeOption(AbstractAttributeOption):
        pass

    __all__.append("AttributeOption")


if not is_model_registered("catalogue", "Option"):

    class Option(AbstractOption):
        pass

    __all__.append("Option")


if not is_model_registered("catalogue", "ProductImage"):

    class ProductImage(AbstractProductImage):
        pass

    __all__.append("ProductImage")


if not is_model_registered("catalogue", "ProductCategoryHierarchy"):

    class ProductCategoryHierarchy(AbstractProductCategoryHierarchy):
        pass

    __all__.append("ProductCategoryHierarchy")



from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig


class PaymentConfig(OscarConfig):
    label = "payment"
    name = "oscar.apps.payment"
    verbose_name = _("Payment")



from django.urls import path, re_path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class UsersDashboardConfig(OscarDashboardConfig):
    label = "users_dashboard"
    name = "oscar.apps.dashboard.users"
    verbose_name = _("Users dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "users-index": DashboardPermission.get("user"),
            "user-detail": DashboardPermission.get("user"),
            "user-password-reset": DashboardPermission.get("user"),
            "user-alert-list": DashboardPermission.get("user"),
            "user-alert-delete": DashboardPermission.get("user"),
            "user-alert-update": DashboardPermission.get("user"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.index_view = get_class("dashboard.users.views", "IndexView")
        self.user_detail_view = get_class("dashboard.users.views", "UserDetailView")
        self.password_reset_view = get_class(
            "dashboard.users.views", "PasswordResetView"
        )
        self.alert_list_view = get_class(
            "dashboard.users.views", "ProductAlertListView"
        )
        self.alert_update_view = get_class(
            "dashboard.users.views", "ProductAlertUpdateView"
        )
        self.alert_delete_view = get_class(
            "dashboard.users.views", "ProductAlertDeleteView"
        )
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path("", self.index_view.as_view(), name="users-index"),
            re_path(
                r"^(?P<pk>-?\d+)/$", self.user_detail_view.as_view(), name="user-detail"
            ),
            re_path(
                r"^(?P<pk>-?\d+)/password-reset/$",
                self.password_reset_view.as_view(),
                name="user-password-reset",
            ),
            # Alerts
            path("alerts/", self.alert_list_view.as_view(), name="user-alert-list"),
            re_path(
                r"^alerts/(?P<pk>-?\d+)/delete/$",
                self.alert_delete_view.as_view(),
                name="user-alert-delete",
            ),
            re_path(
                r"^alerts/(?P<pk>-?\d+)/update/$",
                self.alert_update_view.as_view(),
                name="user-alert-update",
            ),
        ]
        return self.post_process_urls(urls)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('address', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='useraddress',
            old_name='num_orders',
            new_name='num_orders_as_shipping_address',
        ),
        migrations.AddField(
            model_name='useraddress',
            name='num_orders_as_billing_address',
            field=models.PositiveIntegerField(default=0, verbose_name='Number of Orders as Billing Address'),
        ),
    ]



from django.core.exceptions import ImproperlyConfigured
from django.db.models.fields import CharField, DecimalField
from django.utils.translation import gettext_lazy as _
from phonenumber_field.modelfields import PhoneNumberField

from oscar.core import validators
from oscar.forms import fields
from oscar.models.fields.autoslugfield import AutoSlugField


# https://github.com/django/django/blob/64200c14e0072ba0ffef86da46b2ea82fd1e019a/django/db/models/fields/subclassing.py#L31-L44
class Creator(object):
    """
    A placeholder class that provides a way to set the attribute on the model.
    """

    def __init__(self, field):
        self.field = field

    # pylint: disable=W0622
    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return obj.__dict__[self.field.name]

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = self.field.to_python(value)


class ExtendedURLField(CharField):
    description = _("URL")

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = kwargs.get("max_length", 200)
        CharField.__init__(self, *args, **kwargs)
        self.validators.append(validators.ExtendedURLValidator())

    def formfield(self, **kwargs):
        """
        As with CharField, this will cause URL validation to be performed
        twice.
        """
        defaults = {
            "form_class": fields.ExtendedURLField,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def deconstruct(self):
        """
        deconstruct() is needed by Django's migration framework
        """
        name, path, args, kwargs = super().deconstruct()
        # We have a default value for max_length; remove it in that case
        if self.max_length == 200:
            del kwargs["max_length"]
        return name, path, args, kwargs


class PositiveDecimalField(DecimalField):
    """
    A simple subclass of ``django.db.models.fields.DecimalField`` that
    restricts values to be non-negative.
    """

    def formfield(self, **kwargs):
        """
        Return a :py:class:`django.forms.Field` instantiated with a ``min_value`` of 0.
        """
        return super().formfield(min_value=0)


class UppercaseCharField(CharField):
    """
    A simple subclass of ``django.db.models.fields.CharField`` that
    restricts all text to be uppercase.
    """

    def contribute_to_class(self, cls, name, private_only=False, **kwargs):
        super().contribute_to_class(cls, name, private_only=private_only, **kwargs)
        setattr(cls, self.name, Creator(self))

    def from_db_value(self, value, *args, **kwargs):
        return self.to_python(value)

    def to_python(self, value):
        """
        Cast the supplied value to uppercase
        """
        val = super().to_python(value)
        if isinstance(val, str):
            return val.upper()
        else:
            return val


class NullCharField(CharField):
    """
    CharField that stores '' as None and returns None as ''
    Useful when using unique=True and forms. Implies null==blank==True.

    Django's CharField stores '' as None, but does not return None as ''.
    """

    description = "CharField that stores '' as None and returns None as ''"

    def __init__(self, *args, **kwargs):
        if not kwargs.get("null", True) or not kwargs.get("blank", True):
            raise ImproperlyConfigured("NullCharField implies null==blank==True")
        kwargs["null"] = kwargs["blank"] = True
        super().__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, private_only=False, **kwargs):
        super().contribute_to_class(cls, name, private_only=private_only, **kwargs)
        setattr(cls, self.name, Creator(self))

    def from_db_value(self, value, *args, **kwargs):
        value = self.to_python(value)
        # If the value was stored as null, return empty string instead
        return value if value is not None else ""

    def get_prep_value(self, value):
        prepped = super().get_prep_value(value)
        return prepped if prepped != "" else None

    def deconstruct(self):
        """
        deconstruct() is needed by Django's migration framework
        """
        name, path, args, kwargs = super().deconstruct()
        del kwargs["null"]
        del kwargs["blank"]
        return name, path, args, kwargs



from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import ImproperlyConfigured

from oscar.apps.customer.utils import normalise_email
from oscar.core.compat import get_user_model

User = get_user_model()

if hasattr(User, "REQUIRED_FIELDS"):
    if not (User.USERNAME_FIELD == "email" or "email" in User.REQUIRED_FIELDS):
        raise ImproperlyConfigured(
            "EmailBackend: Your User model must have an email field with blank=False"
        )


class EmailBackend(ModelBackend):
    """
    Custom auth backend that uses an email address and password

    For this to work, the User model must have an 'email' field
    """

    # pylint: disable=keyword-arg-before-vararg
    def _authenticate(self, request, email=None, password=None, *args, **kwargs):
        if email is None:
            if "username" not in kwargs or kwargs["username"] is None:
                return None
            clean_email = normalise_email(kwargs["username"])
        else:
            clean_email = normalise_email(email)

        # Check if we're dealing with an email address
        if "@" not in clean_email:
            return None

        # Since Django doesn't enforce emails to be unique, we look for all
        # matching users and try to authenticate them all. Note that we
        # intentionally allow multiple users with the same email address
        # (has been a requirement in larger system deployments),
        # we just enforce that they don't share the same password.
        # We make a case-insensitive match when looking for emails.
        matching_users = User.objects.filter(email__iexact=clean_email)
        authenticated_users = [
            user
            for user in matching_users
            if (user.check_password(password) and self.user_can_authenticate(user))
        ]
        if len(authenticated_users) == 1:
            # Happy path
            return authenticated_users[0]
        elif len(authenticated_users) > 1:
            # This is the problem scenario where we have multiple users with
            # the same email address AND password. We can't safely authenticate
            # either.
            raise User.MultipleObjectsReturned(
                "There are multiple users with the given email address and password"
            )
        return None

    def authenticate(self, *args, **kwargs):
        return self._authenticate(*args, **kwargs)



from django import forms
from django.contrib.auth.models import Permission
from django.contrib.auth.password_validation import validate_password
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.core.compat import existing_user_fields, get_user_model
from oscar.core.loading import get_class, get_model

User = get_user_model()
Partner = get_model("partner", "Partner")
PartnerAddress = get_model("partner", "PartnerAddress")
EmailUserCreationForm = get_class("customer.forms", "EmailUserCreationForm")


class PartnerSearchForm(forms.Form):
    name = forms.CharField(
        required=False, label=pgettext_lazy("Partner's name", "Name")
    )


class PartnerCreateForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Partner.name is optional and that is okay. But if creating through
        # the dashboard, it seems sensible to enforce as it's the only field
        # in the form.
        self.fields["name"].required = True

    class Meta:
        model = Partner
        fields = ("name",)


ROLE_CHOICES = (
    ("staff", _("Full dashboard access")),
    ("limited", _("Limited dashboard access")),
)


class NewUserForm(EmailUserCreationForm):
    role = forms.ChoiceField(
        choices=ROLE_CHOICES,
        widget=forms.RadioSelect,
        label=_("User role"),
        initial="limited",
    )

    def __init__(self, partner, *args, **kwargs):
        self.partner = partner
        super().__init__(host=None, *args, **kwargs)

    def save(self):
        role = self.cleaned_data.get("role", "limited")
        user = super().save(commit=False)
        user.is_staff = role == "staff"
        user.save()
        self.partner.users.add(user)
        if role == "limited":
            dashboard_access_perm = Permission.objects.get(
                codename="dashboard_access", content_type__app_label="partner"
            )
            user.user_permissions.add(dashboard_access_perm)
        return user

    class Meta:
        model = User
        fields = existing_user_fields(["first_name", "last_name", "email"]) + [
            "password1",
            "password2",
        ]


class ExistingUserForm(forms.ModelForm):
    """
    Slightly different form that makes
    * makes saving password optional
    * doesn't regenerate username
    * doesn't allow changing email till #668 is resolved
    """

    role = forms.ChoiceField(
        choices=ROLE_CHOICES, widget=forms.RadioSelect, label=_("User role")
    )
    password1 = forms.CharField(
        label=_("Password"), widget=forms.PasswordInput, required=False
    )
    password2 = forms.CharField(
        required=False, label=_("Confirm Password"), widget=forms.PasswordInput
    )

    def clean_password2(self):
        password1 = self.cleaned_data.get("password1", "")
        password2 = self.cleaned_data.get("password2", "")

        if password1 != password2:
            raise forms.ValidationError(_("The two password fields didn't match."))
        validate_password(password2, self.instance)
        return password2

    def __init__(self, *args, **kwargs):
        user = kwargs["instance"]
        role = "staff" if user.is_staff else "limited"
        kwargs.get("initial", {}).setdefault("role", role)
        super().__init__(*args, **kwargs)

    def save(self, commit=False):
        role = self.cleaned_data.get("role", "none")
        user = super().save(commit=False)
        user.is_staff = role == "staff"
        if self.cleaned_data["password1"]:
            user.set_password(self.cleaned_data["password1"])
        user.save()

        dashboard_perm = Permission.objects.get(
            codename="dashboard_access", content_type__app_label="partner"
        )
        user_has_perm = user.user_permissions.filter(pk=dashboard_perm.pk).exists()
        if role == "limited" and not user_has_perm:
            user.user_permissions.add(dashboard_perm)
        elif role == "staff" and user_has_perm:
            user.user_permissions.remove(dashboard_perm)
        return user

    class Meta:
        model = User
        fields = existing_user_fields(["first_name", "last_name"]) + [
            "password1",
            "password2",
        ]


class UserEmailForm(forms.Form):
    # We use a CharField so that a partial email address can be entered
    email = forms.CharField(label=_("Email address"), max_length=100)


class PartnerAddressForm(forms.ModelForm):
    name = forms.CharField(
        required=False, max_length=128, label=pgettext_lazy("Partner's name", "Name")
    )

    class Meta:
        fields = (
            "name",
            "line1",
            "line2",
            "line3",
            "line4",
            "state",
            "postcode",
            "country",
        )
        model = PartnerAddress



from django.conf import settings
from django.contrib import messages
from django.db.models import Count, Q
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.views.generic import DeleteView, ListView

from oscar.core.loading import get_class, get_classes, get_model
from oscar.views import sort_queryset

ConditionalOffer = get_model("offer", "ConditionalOffer")
Condition = get_model("offer", "Condition")
Range = get_model("offer", "Range")
Product = get_model("catalogue", "Product")
OrderDiscount = get_model("order", "OrderDiscount")
Benefit = get_model("offer", "Benefit")
(
    MetaDataForm,
    ConditionForm,
    BenefitForm,
    RestrictionsForm,
    OfferSearchForm,
) = get_classes(
    "dashboard.offers.forms",
    [
        "MetaDataForm",
        "ConditionForm",
        "BenefitForm",
        "RestrictionsForm",
        "OfferSearchForm",
    ],
)
OfferWizardStepView = get_class("dashboard.offers.wizard_views", "OfferWizardStepView")
OrderDiscountCSVFormatter = get_class(
    "dashboard.offers.reports", "OrderDiscountCSVFormatter"
)


# pylint: disable=attribute-defined-outside-init
class OfferListView(ListView):
    model = ConditionalOffer
    context_object_name = "offers"
    template_name = "oscar/dashboard/offers/offer_list.html"
    form_class = OfferSearchForm
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE

    def get_queryset(self):
        self.search_filters = []
        qs = self.model._default_manager.annotate(
            voucher_count=Count("vouchers")
        ).select_related("benefit", "condition")
        qs = sort_queryset(
            qs,
            self.request,
            [
                "name",
                "offer_type",
                "start_datetime",
                "end_datetime",
                "num_applications",
                "total_discount",
            ],
        )

        self.form = self.form_class(self.request.GET)
        # This form is exactly the same as the other one, apart from having
        # fields with different IDs, so that they are unique within the page
        # (non-unique field IDs also break Select2)
        self.advanced_form = self.form_class(self.request.GET, auto_id="id_advanced_%s")
        if not all([self.form.is_valid(), self.advanced_form.is_valid()]):
            return qs

        name = self.form.cleaned_data["name"]
        offer_type = self.form.cleaned_data["offer_type"]
        is_active = self.form.cleaned_data["is_active"]
        has_vouchers = self.form.cleaned_data["has_vouchers"]
        voucher_code = self.form.cleaned_data["voucher_code"]

        if name:
            qs = qs.filter(name__icontains=name)
            self.search_filters.append(_('Name matches "%s"') % name)
        if is_active is not None:
            now = timezone.now()
            if is_active:
                qs = qs.filter(
                    (Q(start_datetime__lte=now) | Q(start_datetime__isnull=True))
                    & (Q(end_datetime__gte=now) | Q(end_datetime__isnull=True)),
                    status=ConditionalOffer.OPEN,
                )
                self.search_filters.append(_("Is active"))
            else:
                qs = qs.filter(
                    Q(start_datetime__gt=now)
                    | Q(end_datetime__lt=now)
                    | Q(status=ConditionalOffer.SUSPENDED)
                )
                self.search_filters.append(_("Is inactive"))
        if offer_type:
            qs = qs.filter(offer_type=offer_type)
            self.search_filters.append(
                _('Is of type "%s"') % dict(ConditionalOffer.TYPE_CHOICES)[offer_type]
            )
        if has_vouchers is not None:
            qs = qs.filter(vouchers__isnull=not has_vouchers).distinct()
            self.search_filters.append(
                _("Has vouchers") if has_vouchers else _("Has no vouchers")
            )
        if voucher_code:
            qs = qs.filter(vouchers__code__icontains=voucher_code).distinct()
            self.search_filters.append(_('Voucher code matches "%s"') % voucher_code)

        return qs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["form"] = self.form
        ctx["advanced_form"] = self.advanced_form
        ctx["search_filters"] = self.search_filters
        return ctx


class OfferMetaDataView(OfferWizardStepView):
    step_name = "metadata"
    form_class = MetaDataForm
    template_name = "oscar/dashboard/offers/metadata_form.html"
    url_name = "dashboard:offer-metadata"
    success_url_name = "dashboard:offer-benefit"

    def get_instance(self):
        return self.offer

    def get_title(self):
        return _("Name, description and type")


class OfferBenefitView(OfferWizardStepView):
    step_name = "benefit"
    form_class = BenefitForm
    template_name = "oscar/dashboard/offers/benefit_form.html"
    url_name = "dashboard:offer-benefit"
    success_url_name = "dashboard:offer-condition"
    previous_view = OfferMetaDataView

    def get_instance(self):
        return self.offer.benefit

    def get_title(self):
        # This is referred to as the 'incentive' within the dashboard.
        return _("Incentive")


class OfferConditionView(OfferWizardStepView):
    step_name = "condition"
    form_class = ConditionForm
    template_name = "oscar/dashboard/offers/condition_form.html"
    url_name = "dashboard:offer-condition"
    success_url_name = "dashboard:offer-restrictions"
    previous_view = OfferBenefitView

    def get_instance(self):
        return self.offer.condition


class OfferRestrictionsView(OfferWizardStepView):
    step_name = "restrictions"
    form_class = RestrictionsForm
    template_name = "oscar/dashboard/offers/restrictions_form.html"
    previous_view = OfferConditionView
    url_name = "dashboard:offer-restrictions"

    def form_valid(self, form):
        offer = form.save(commit=False)
        return self.save_offer(offer)

    def get_instance(self):
        return self.offer

    def get_title(self):
        return _("Restrictions")


class OfferDeleteView(DeleteView):
    model = ConditionalOffer
    template_name = "oscar/dashboard/offers/offer_delete.html"
    context_object_name = "offer"

    def dispatch(self, request, *args, **kwargs):
        offer = self.get_object()
        if offer.vouchers.exists():
            messages.warning(
                request,
                _(
                    "This offer can only be deleted if it has no vouchers attached to it"
                ),
            )
            return redirect("dashboard:offer-detail", pk=offer.pk)
        return super().dispatch(request, *args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, _("Offer deleted!"))
        return reverse("dashboard:offer-list")


class OfferDetailView(ListView):
    # Slightly odd, but we treat the offer detail view as a list view so the
    # order discounts can be browsed.
    model = OrderDiscount
    template_name = "oscar/dashboard/offers/offer_detail.html"
    context_object_name = "order_discounts"
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE

    # pylint: disable=attribute-defined-outside-init
    def dispatch(self, request, *args, **kwargs):
        self.offer = get_object_or_404(ConditionalOffer, pk=kwargs["pk"])
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        if "suspend" in request.POST:
            return self.suspend()
        elif "unsuspend" in request.POST:
            return self.unsuspend()

    def suspend(self):
        if self.offer.is_suspended:
            messages.error(self.request, _("Offer is already suspended"))
        else:
            self.offer.suspend()
            messages.success(self.request, _("Offer suspended"))
        return HttpResponseRedirect(
            reverse("dashboard:offer-detail", kwargs={"pk": self.offer.pk})
        )

    def unsuspend(self):
        if not self.offer.is_suspended:
            messages.error(
                self.request,
                _("Offer cannot be reinstated as it is not currently suspended"),
            )
        else:
            self.offer.unsuspend()
            messages.success(self.request, _("Offer reinstated"))
        return HttpResponseRedirect(
            reverse("dashboard:offer-detail", kwargs={"pk": self.offer.pk})
        )

    def get_queryset(self):
        return self.model.objects.filter(offer_id=self.offer.pk).select_related("order")

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx["offer"] = self.offer
        return ctx

    def render_to_response(self, context, *args, **kwargs):
        if self.request.GET.get("format") == "csv":
            formatter = OrderDiscountCSVFormatter()
            return formatter.generate_response(self.get_queryset(), offer=self.offer)
        return super().render_to_response(context, *args, **kwargs)



from django.db import models
from django.utils.timezone import now

from .queryset import RangeQuerySet


class ActiveOfferManager(models.Manager):
    """
    For searching/creating offers within their date range
    """

    def get_queryset(self):
        cutoff = now()
        return (
            super()
            .get_queryset()
            .filter(
                models.Q(end_datetime__gte=cutoff) | models.Q(end_datetime=None),
                models.Q(start_datetime__lte=cutoff) | models.Q(start_datetime=None),
            )
            .filter(status=self.model.OPEN)
        )


RangeManager = models.Manager.from_queryset(RangeQuerySet, "RangeManager")


class BrowsableRangeManager(RangeManager):
    """
    For searching only ranges which have the "is_public" flag set to True.
    """

    def get_queryset(self):
        return super().get_queryset().filter(is_public=True)



from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class PartnersDashboardConfig(OscarDashboardConfig):
    label = "partners_dashboard"
    name = "oscar.apps.dashboard.partners"
    verbose_name = _("Partners dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "partner-list": DashboardPermission.get("partner"),
            "partner-create": DashboardPermission.get("partner"),
            "partner-manage": DashboardPermission.get("partner"),
            "partner-delete": DashboardPermission.get("partner"),
            "partner-user-create": DashboardPermission.get("partner"),
            "partner-user-select": DashboardPermission.get("partner"),
            "partner-user-link": DashboardPermission.get("partner"),
            "partner-user-unlink": DashboardPermission.get("partner"),
            "partner-user-update": DashboardPermission.get("partner"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.list_view = get_class("dashboard.partners.views", "PartnerListView")
        self.create_view = get_class("dashboard.partners.views", "PartnerCreateView")
        self.manage_view = get_class("dashboard.partners.views", "PartnerManageView")
        self.delete_view = get_class("dashboard.partners.views", "PartnerDeleteView")

        self.user_link_view = get_class(
            "dashboard.partners.views", "PartnerUserLinkView"
        )
        self.user_unlink_view = get_class(
            "dashboard.partners.views", "PartnerUserUnlinkView"
        )
        self.user_create_view = get_class(
            "dashboard.partners.views", "PartnerUserCreateView"
        )
        self.user_select_view = get_class(
            "dashboard.partners.views", "PartnerUserSelectView"
        )
        self.user_update_view = get_class(
            "dashboard.partners.views", "PartnerUserUpdateView"
        )
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path("", self.list_view.as_view(), name="partner-list"),
            path("create/", self.create_view.as_view(), name="partner-create"),
            path("<int:pk>/", self.manage_view.as_view(), name="partner-manage"),
            path("<int:pk>/delete/", self.delete_view.as_view(), name="partner-delete"),
            path(
                "<int:partner_pk>/users/add/",
                self.user_create_view.as_view(),
                name="partner-user-create",
            ),
            path(
                "<int:partner_pk>/users/select/",
                self.user_select_view.as_view(),
                name="partner-user-select",
            ),
            path(
                "<int:partner_pk>/users/<int:user_pk>/link/",
                self.user_link_view.as_view(),
                name="partner-user-link",
            ),
            path(
                "<int:partner_pk>/users/<int:user_pk>/unlink/",
                self.user_unlink_view.as_view(),
                name="partner-user-unlink",
            ),
            path(
                "<int:partner_pk>/users/<int:user_pk>/update/",
                self.user_update_view.as_view(),
                name="partner-user-update",
            ),
        ]
        return self.post_process_urls(urls)



# Generated by Django 3.2.19 on 2023-07-10 08:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0013_json_option_value'),
    ]

    operations = [
        migrations.AddField(
            model_name='line',
            name='tax_code',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='VAT rate code'),
        ),
        migrations.AddField(
            model_name='lineprice',
            name='tax_code',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='VAT rate code'),
        ),
        migrations.AddField(
            model_name='order',
            name='shipping_tax_code',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='Shipping VAT rate code'),
        ),
        migrations.AddField(
            model_name='surcharge',
            name='tax_code',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='VAT rate code'),
        ),
    ]



# pylint: disable=E1101
from urllib.parse import quote

from django.contrib import messages
from django.http import Http404, HttpResponsePermanentRedirect
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

BrowseCategoryForm = get_class("search.forms", "BrowseCategoryForm")
CategoryForm = get_class("search.forms", "CategoryForm")
BaseSearchView = get_class("search.views.base", "BaseSearchView")
Category = get_model("catalogue", "Category")


class CatalogueView(BaseSearchView):
    """
    Browse all products in the catalogue
    """

    form_class = BrowseCategoryForm
    context_object_name = "products"
    template_name = "oscar/catalogue/browse.html"
    enforce_paths = True

    def get(self, request, *args, **kwargs):
        try:
            return super().get(request, *args, **kwargs)
        except Http404:
            # Redirect to page one.
            messages.error(request, _("The given page number was invalid."))
            return redirect("catalogue:index")

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx["summary"] = _("All products")
        return ctx


class ProductCategoryView(BaseSearchView):
    """
    Browse products in a given category
    """

    form_class = CategoryForm
    enforce_paths = True
    context_object_name = "products"
    template_name = "oscar/catalogue/category.html"

    def get(self, request, *args, **kwargs):
        # pylint: disable=W0201
        self.category = self.get_category()

        # Allow staff members so they can test layout etc.
        if not self.is_viewable(self.category, request):
            raise Http404()

        potential_redirect = self.redirect_if_necessary(request.path, self.category)
        if potential_redirect is not None:
            return potential_redirect

        try:
            return super().get(request, *args, **kwargs)
        except Http404:
            messages.error(request, _("The given page number was invalid."))
            return redirect(self.category.get_absolute_url())

    def is_viewable(self, category, request):
        return category.is_public or request.user.is_staff

    def redirect_if_necessary(self, current_path, category):
        if self.enforce_paths:
            # Categories are fetched by primary key to allow slug changes.
            # If the slug has changed, issue a redirect.
            expected_path = category.get_absolute_url()
            if expected_path != quote(current_path):
                return HttpResponsePermanentRedirect(expected_path)

    def get_category(self):
        return get_object_or_404(Category, pk=self.kwargs["pk"])

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["category"] = self.category
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["categories"] = self.category.get_descendants_and_self()
        return kwargs



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0003_data_migration_slugs'),
    ]

    operations = [
        migrations.AlterField(
            model_name='category',
            name='slug',
            field=models.SlugField(max_length=255, verbose_name='Slug'),
            preserve_default=True,
        ),
    ]



import factory
from phonenumber_field.phonenumber import PhoneNumber

from oscar.core.loading import get_model

__all__ = [
    "CountryFactory",
    "UserAddressFactory",
]


class CountryFactory(factory.django.DjangoModelFactory):
    iso_3166_1_a2 = "GB"
    printable_name = "UNITED KINGDOM"

    class Meta:
        model = get_model("address", "Country")
        django_get_or_create = ("iso_3166_1_a2",)


class UserAddressFactory(factory.django.DjangoModelFactory):
    title = "Dr"
    first_name = "Barry"
    last_name = "Barrington"
    line1 = "1 King Road"
    line4 = "London"
    postcode = "SW1 9RE"
    phone_number = PhoneNumber.from_string("+49 351 3296645")
    country = factory.SubFactory(CountryFactory)
    user = factory.SubFactory("oscar.test.factories.UserFactory")

    class Meta:
        model = get_model("address", "UserAddress")



from oscar.apps.basket.abstract_models import (
    AbstractBasket,
    AbstractLine,
    AbstractLineAttribute,
)
from oscar.core.loading import is_model_registered

__all__ = [
    "InvalidBasketLineError",
]


class InvalidBasketLineError(Exception):
    pass


if not is_model_registered("basket", "Basket"):

    class Basket(AbstractBasket):
        pass

    __all__.append("Basket")


if not is_model_registered("basket", "Line"):

    class Line(AbstractLine):
        pass

    __all__.append("Line")


if not is_model_registered("basket", "LineAttribute"):

    class LineAttribute(AbstractLineAttribute):
        pass

    __all__.append("LineAttribute")



class PaymentError(Exception):
    pass


class UserCancelled(PaymentError):
    """
    Exception for when a customer decides to cancel their payment
    after the process has started -- for example if they press a "Cancel"
    button on a third-party payment platform.
    """


class TransactionDeclined(PaymentError):
    pass


class GatewayError(PaymentError):
    pass


class InvalidGatewayRequestError(PaymentError):
    pass


class InsufficientPaymentSources(PaymentError):
    """
    Exception for when a user attempts to checkout without specifying enough
    payment sources to cover the entire order total.

    Eg. When selecting an allocation off a giftcard but not specifying a
    bankcard to take the remainder from.
    """


class RedirectRequired(PaymentError):
    """
    Exception to be used when payment processsing requires a redirect
    """

    def __init__(self, url):
        self.url = url


class UnableToTakePayment(PaymentError):
    """
    Exception to be used for ANTICIPATED payment errors (eg card number wrong,
    expiry date has passed).  The message passed here will be shown to the end
    user.
    """



# pylint: disable=attribute-defined-outside-init
import datetime

from django.conf import settings
from django.db.models import Q
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views import generic

from oscar.core.loading import get_classes, get_model
from oscar.core.utils import format_datetime
from oscar.views import sort_queryset
from oscar.views.generic import BulkEditMixin

ProductReviewSearchForm, DashboardProductReviewForm = get_classes(
    "dashboard.reviews.forms", ("ProductReviewSearchForm", "DashboardProductReviewForm")
)
ProductReview = get_model("reviews", "productreview")


class ReviewListView(BulkEditMixin, generic.ListView):
    model = ProductReview
    template_name = "oscar/dashboard/reviews/review_list.html"
    context_object_name = "review_list"
    form_class = ProductReviewSearchForm
    review_form_class = DashboardProductReviewForm
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE
    actions = ("update_selected_review_status",)
    checkbox_object_name = "review"
    desc_template = _(
        "%(main_filter)s %(date_filter)s %(status_filter)s"
        "%(kw_filter)s %(name_filter)s"
    )

    def get(self, request, *args, **kwargs):
        response = super().get(request, **kwargs)
        self.form = self.form_class()
        return response

    def get_date_from_to_queryset(self, date_from, date_to, queryset=None):
        """
        Get a ``QuerySet`` of ``ProductReview`` items that match the time
        frame specified by *date_from* and *date_to*. Both parameters are
        expected to be in ``datetime`` format with *date_from* < *date_to*.
        If *queryset* is specified, it will be filtered according to the
        given dates. Otherwise, a new queryset for all ``ProductReview``
        items is created.
        """
        if queryset is None:
            queryset = self.model.objects.all()

        if date_from:
            queryset = queryset.filter(date_created__gte=date_from)
            self.desc_ctx["date_filter"] = _(" created after %s") % format_datetime(
                date_from
            )
        if date_to:
            # Add 24 hours to make search inclusive
            date_to = date_to + datetime.timedelta(days=1)
            queryset = queryset.filter(date_created__lt=date_to)
            self.desc_ctx["date_filter"] = _(" created before %s") % format_datetime(
                date_to
            )

        if date_from and date_to:
            # override description
            self.desc_ctx["date_filter"] = _(
                " created between %(start_date)s and %(end_date)s"
            ) % {
                "start_date": format_datetime(date_from),
                "end_date": format_datetime(date_to),
            }
        return queryset

    def get_queryset(self):
        queryset = self.model.objects.select_related("product", "user").all()
        queryset = sort_queryset(
            queryset,
            self.request,
            ["date_created", "score", "total_votes"],
            default="-date_created",
        )
        self.desc_ctx = {
            "main_filter": _("All reviews"),
            "date_filter": "",
            "status_filter": "",
            "kw_filter": "",
            "name_filter": "",
        }

        self.form = self.form_class(self.request.GET)
        if not self.form.is_valid():
            return queryset

        data = self.form.cleaned_data

        queryset = self.add_filter_status(queryset, data["status"])
        queryset = self.add_filter_keyword(queryset, data["keyword"])
        queryset = self.add_filter_name(queryset, data["name"])

        queryset = self.get_date_from_to_queryset(
            data["date_from"], data["date_to"], queryset
        )

        return queryset

    def add_filter_status(self, queryset, status):
        # checking for empty string rather then True is required
        # as zero is a valid value for 'status' but would be
        # evaluated to False
        if status != "":
            queryset = queryset.filter(status=status).distinct()
            display_status = self.form.get_friendly_status()
            self.desc_ctx["status_filter"] = (
                _(" with status matching '%s'") % display_status
            )
        return queryset

    def add_filter_keyword(self, queryset, keyword):
        if keyword:
            queryset = queryset.filter(
                Q(title__icontains=keyword) | Q(body__icontains=keyword)
            ).distinct()
            self.desc_ctx["kw_filter"] = _(" with keyword matching '%s'") % keyword
        return queryset

    def add_filter_name(self, queryset, name):
        if name:
            # If the value is two words, then assume they are first name and
            # last name
            parts = name.split()
            if len(parts) >= 2:
                queryset = queryset.filter(
                    user__first_name__istartswith=parts[0],
                    user__last_name__istartswith=parts[1],
                ).distinct()
            else:
                queryset = queryset.filter(
                    Q(user__first_name__istartswith=parts[0])
                    | Q(user__last_name__istartswith=parts[-1])
                ).distinct()
            self.desc_ctx["name_filter"] = _(" with customer name matching '%s'") % name

        return queryset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["review_form"] = self.review_form_class()
        context["form"] = self.form
        context["description"] = self.desc_template % self.desc_ctx
        return context

    def update_selected_review_status(self, request, reviews):
        """
        Update the status of the selected *reviews* with the new
        status in the *request* POST data. Redirects back to the
        list view of reviews.
        """
        new_status = int(request.POST.get("status"))
        for review in reviews:
            review.status = new_status
            review.save()
        return HttpResponseRedirect(reverse("dashboard:reviews-list"))


class ReviewUpdateView(generic.UpdateView):
    model = ProductReview
    template_name = "oscar/dashboard/reviews/review_update.html"
    form_class = DashboardProductReviewForm
    context_object_name = "review"

    def get_success_url(self):
        return reverse("dashboard:reviews-list")


class ReviewDeleteView(generic.DeleteView):
    model = ProductReview
    template_name = "oscar/dashboard/reviews/review_delete.html"
    context_object_name = "review"

    def get_success_url(self):
        return reverse("dashboard:reviews-list")



import datetime

import factory
from django.utils.timezone import now

from oscar.apps.voucher.utils import get_unused_code
from oscar.core.loading import get_model
from oscar.test.factories import ConditionalOfferFactory

ConditionalOffer = get_model("offer", "ConditionalOffer")
Voucher = get_model("voucher", "Voucher")

__all__ = ["VoucherFactory", "VoucherSetFactory"]


class VoucherFactory(factory.django.DjangoModelFactory):
    name = "My voucher"
    code = "MYVOUCHER"
    usage = Voucher.MULTI_USE

    start_datetime = now() - datetime.timedelta(days=1)
    end_datetime = now() + datetime.timedelta(days=10)

    class Meta:
        model = get_model("voucher", "Voucher")


class VoucherSetFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: "Voucher Set %d" % n)
    count = 100
    code_length = 12
    description = "Dummy description"
    start_datetime = now() - datetime.timedelta(days=1)
    end_datetime = now() + datetime.timedelta(days=10)

    class Meta:
        model = get_model("voucher", "VoucherSet")

    # pylint: disable=unused-argument
    @factory.post_generation
    def vouchers(obj, create, extracted, **kwargs):
        if not create:
            return
        offer = ConditionalOfferFactory(offer_type=ConditionalOffer.VOUCHER)
        for i in range(0, obj.count):
            voucher = Voucher.objects.create(
                name="%s - %d" % (obj.name, i + 1),
                code=get_unused_code(length=obj.code_length),
                voucher_set=obj,
                usage=Voucher.MULTI_USE,
                start_datetime=obj.start_datetime,
                end_datetime=obj.end_datetime,
            )
            voucher.offers.add(offer)



# Generated by Django 3.2.9 on 2022-03-28 08:39

from django.db import migrations, models
import django.db.models.deletion

from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)


class Migration(migrations.Migration):

    dependencies = [
        ('wishlists', '0003_auto_20181115_1953'),
    ]

    operations = [
        migrations.CreateModel(
            name='WishListSharedEmail',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=254, verbose_name='Email')),
                ('wishlist', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='shared_emails', to='wishlists.wishlist', verbose_name='Wish List')),
            ],
            options={
                'verbose_name': 'Wish list shared email',
                'abstract': False,
            },
        ),
    ]



from django.urls import reverse_lazy

from oscar.forms.widgets import MultipleRemoteSelect, RemoteSelect


class ProductSelect(RemoteSelect):
    # Implemented as separate class instead of just calling
    # AjaxSelect(data_url=...) for overridability and backwards compatibility
    lookup_url = reverse_lazy("dashboard:catalogue-product-lookup")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attrs["class"] = "select2 product-select"


class ProductSelectMultiple(MultipleRemoteSelect):
    # Implemented as separate class instead of just calling
    # AjaxSelect(data_url=...) for overridability and backwards compatibility
    lookup_url = reverse_lazy("dashboard:catalogue-product-lookup")



from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

ReportGenerator = get_class("dashboard.reports.reports", "ReportGenerator")
ReportCSVFormatter = get_class("dashboard.reports.reports", "ReportCSVFormatter")
ReportHTMLFormatter = get_class("dashboard.reports.reports", "ReportHTMLFormatter")
Order = get_model("order", "Order")


class OrderReportCSVFormatter(ReportCSVFormatter):
    filename_template = "orders-%s-to-%s.csv"

    def generate_csv(self, response, orders):
        writer = self.get_csv_writer(response)
        header_row = [
            _("Order number"),
            _("Name"),
            _("Email"),
            _("Total incl. tax"),
            _("Date placed"),
        ]
        writer.writerow(header_row)
        for order in orders:
            row = [
                order.number,
                "-" if order.is_anonymous else order.user.get_full_name(),
                order.email,
                order.total_incl_tax,
                self.format_datetime(order.date_placed),
            ]
            writer.writerow(row)

    def filename(self, **kwargs):
        return self.filename_template % (kwargs["start_date"], kwargs["end_date"])


class OrderReportHTMLFormatter(ReportHTMLFormatter):
    filename_template = "oscar/dashboard/reports/partials/order_report.html"


class OrderReportGenerator(ReportGenerator):
    code = "order_report"
    description = _("Orders placed")
    date_range_field_name = "date_placed"
    model_class = Order

    formatters = {
        "CSV_formatter": OrderReportCSVFormatter,
        "HTML_formatter": OrderReportHTMLFormatter,
    }

    def generate(self):
        additional_data = {"start_date": self.start_date, "end_date": self.end_date}
        return self.formatter.generate_response(self.queryset, **additional_data)

    def is_available_to(self, user):
        return user.is_staff



# pylint: disable=unused-argument
from decimal import Decimal as D

from django.conf import settings
from django.contrib.sites.models import Site
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from oscar.apps.order.signals import order_placed
from oscar.core.loading import get_class, get_model

from . import exceptions

Order = get_model("order", "Order")
Line = get_model("order", "Line")
OrderDiscount = get_model("order", "OrderDiscount")
OrderLineDiscount = get_model("order", "OrderLineDiscount")
CommunicationEvent = get_model("order", "CommunicationEvent")
CommunicationEventType = get_model("communication", "CommunicationEventType")
Dispatcher = get_class("communication.utils", "Dispatcher")
Surcharge = get_model("order", "Surcharge")


class OrderNumberGenerator(object):
    """
    Simple object for generating order numbers.

    We need this as the order number is often required for payment
    which takes place before the order model has been created.
    """

    def order_number(self, basket):
        """
        Return an order number for a given basket
        """
        return 100000 + basket.id


class OrderCreator(object):
    """
    Places the order by writing out the various models
    """

    def place_order(
        self,
        basket,
        total,
        shipping_method,
        shipping_charge,
        user=None,
        shipping_address=None,
        billing_address=None,
        order_number=None,
        status=None,
        request=None,
        surcharges=None,
        **kwargs,
    ):
        """
        Placing an order involves creating all the relevant models based on the
        basket and session data.
        """
        if basket.is_empty:
            raise ValueError(_("Empty baskets cannot be submitted"))
        if not order_number:
            generator = OrderNumberGenerator()
            order_number = generator.order_number(basket)
        if not status and hasattr(settings, "OSCAR_INITIAL_ORDER_STATUS"):
            status = getattr(settings, "OSCAR_INITIAL_ORDER_STATUS")

        if Order._default_manager.filter(number=order_number).exists():
            raise ValueError(
                _("There is already an order with number %s") % order_number
            )

        with transaction.atomic():
            kwargs["surcharges"] = surcharges
            # Ok - everything seems to be in order, let's place the order
            order = self.create_order_model(
                user,
                basket,
                shipping_address,
                shipping_method,
                shipping_charge,
                billing_address,
                total,
                order_number,
                status,
                request,
                **kwargs,
            )
            for voucher in basket.vouchers.select_for_update():
                if not voucher.is_active():  # basket ignores inactive vouchers
                    basket.vouchers.remove(voucher)
                else:
                    available_to_user, msg = voucher.is_available_to_user(user=user)
                    if not available_to_user:
                        raise ValueError(msg)

            # Record any discounts associated with this order
            for application in basket.offer_applications:
                # Trigger any deferred benefits from offers and capture the
                # resulting message
                application["message"] = application["offer"].apply_deferred_benefit(
                    basket, order, application
                )
                # Record offer application results
                if application["result"].affects_shipping:
                    # Skip zero shipping discounts
                    shipping_discount = shipping_method.discount(basket)
                    if shipping_discount <= D("0.00"):
                        continue
                    # If a shipping offer, we need to grab the actual discount off
                    # the shipping method instance, which should be wrapped in an
                    # OfferDiscount instance.
                    application["discount"] = shipping_discount
                self.create_discount_model(order, application)
                self.record_discount(application)

            for voucher in basket.vouchers.all():
                self.record_voucher_usage(order, voucher, user)

            for line in basket.all_lines():
                self.create_line_models(order, line)
                self.update_stock_records(line)

        # Send signal for analytics to pick up
        order_placed.send(sender=self, order=order, user=user)

        return order

    def create_order_model(
        self,
        user,
        basket,
        shipping_address,
        shipping_method,
        shipping_charge,
        billing_address,
        total,
        order_number,
        status,
        request=None,
        surcharges=None,
        **extra_order_fields,
    ):
        """Create an order model."""
        order_data = {
            "basket": basket,
            "number": order_number,
            "currency": total.currency,
            "total_incl_tax": total.incl_tax,
            "total_excl_tax": total.excl_tax,
            "shipping_incl_tax": shipping_charge.incl_tax,
            "shipping_excl_tax": shipping_charge.excl_tax,
            "shipping_tax_code": shipping_charge.tax_code,
            "shipping_method": shipping_method.name,
            "shipping_code": shipping_method.code,
        }
        if shipping_address:
            order_data["shipping_address"] = shipping_address
        if billing_address:
            order_data["billing_address"] = billing_address
        if user and user.is_authenticated:
            order_data["user_id"] = user.id
        if status:
            order_data["status"] = status
        if extra_order_fields:
            order_data.update(extra_order_fields)
        if "site" not in order_data:
            order_data["site"] = Site._default_manager.get_current(request)
        order = Order(**order_data)
        order.save()
        if surcharges is not None:
            for charge in surcharges:
                Surcharge.objects.create(
                    order=order,
                    name=charge.surcharge.name,
                    code=charge.surcharge.code,
                    excl_tax=charge.price.excl_tax,
                    incl_tax=charge.price.incl_tax,
                    tax_code=charge.price.tax_code,
                )

        return order

    def create_line_models(self, order, basket_line, extra_line_fields=None):
        """
        Create the batch line model.

        You can set extra fields by passing a dictionary as the
        extra_line_fields value
        """
        product = basket_line.product
        stockrecord = basket_line.stockrecord
        if not stockrecord:
            raise exceptions.UnableToPlaceOrder(
                "Basket line #%d has no stockrecord" % basket_line.id
            )

        partner = stockrecord.partner
        line_data = {
            "order": order,
            # Partner details
            "partner": partner,
            "partner_name": partner.name,
            "partner_sku": stockrecord.partner_sku,
            "stockrecord": stockrecord,
            # Product details
            "product": product,
            "title": product.get_title(),
            "upc": product.upc,
            "quantity": basket_line.quantity,
            # Price details
            "line_price_excl_tax": basket_line.line_price_excl_tax_incl_discounts,
            "line_price_incl_tax": basket_line.line_price_incl_tax_incl_discounts,
            "line_price_before_discounts_excl_tax": basket_line.line_price_excl_tax,
            "line_price_before_discounts_incl_tax": basket_line.line_price_incl_tax,
            # Reporting details
            "unit_price_incl_tax": basket_line.unit_price_incl_tax,
            "unit_price_excl_tax": basket_line.unit_price_excl_tax,
            "tax_code": basket_line.tax_code,
            "num_allocated": basket_line.quantity,
        }
        extra_line_fields = extra_line_fields or {}
        if hasattr(settings, "OSCAR_INITIAL_LINE_STATUS"):
            if not (extra_line_fields and "status" in extra_line_fields):
                extra_line_fields["status"] = getattr(
                    settings, "OSCAR_INITIAL_LINE_STATUS"
                )
        if extra_line_fields:
            line_data.update(extra_line_fields)

        order_line = Line._default_manager.create(**line_data)
        self.create_line_price_models(order, order_line, basket_line)
        self.create_line_attributes(order, order_line, basket_line)
        self.create_line_discount_models(order, order_line, basket_line)
        self.create_additional_line_models(order, order_line, basket_line)

        return order_line

    def update_stock_records(self, line):
        """
        Update any relevant stock records for this order line
        """
        if line.product.get_product_class().track_stock:
            line.stockrecord.allocate(line.quantity)

    def create_line_discount_models(self, order, order_line, basket_line):
        for discount in basket_line.discounts:
            if not discount.offer:
                continue
            order_discount = order.discounts.filter(offer_id=discount.offer.id).first()
            # If we are unable to find the discount we do not care, the total amount is still saved on the discount model,
            # these models are only created so we know how much discount was given per line for each offer
            if order_discount:
                order_line.discounts.create(
                    order_discount=order_discount,
                    is_incl_tax=discount.incl_tax,
                    amount=discount.amount,
                )

    def create_additional_line_models(self, order, order_line, basket_line):
        """
        Empty method designed to be overridden.

        Some applications require additional information about lines, this
        method provides a clean place to create additional models that
        relate to a given line.
        """
        return

    def create_line_price_models(self, order, order_line, basket_line):
        """
        Creates the batch line price models
        """
        breakdown = basket_line.get_price_breakdown()
        for price_incl_tax, price_excl_tax, quantity in breakdown:
            order_line.prices.create(
                order=order,
                quantity=quantity,
                price_incl_tax=price_incl_tax,
                price_excl_tax=price_excl_tax,
                tax_code=basket_line.tax_code,
            )

    # pylint: disable=unused-argument
    def create_line_attributes(self, order, order_line, basket_line):
        """
        Creates the batch line attributes.
        """
        for attr in basket_line.attributes.all():
            order_line.attributes.create(
                option=attr.option, type=attr.option.code, value=attr.value
            )

    def create_discount_model(self, order, discount):
        """
        Create an order discount model for each offer application attached to
        the basket.
        """
        order_discount = OrderDiscount(
            order=order,
            message=discount["message"] or "",
            offer_id=discount["offer"].id,
            frequency=discount["freq"],
            amount=discount["discount"],
        )
        result = discount["result"]
        if result.affects_shipping:
            order_discount.category = OrderDiscount.SHIPPING
        elif result.affects_post_order:
            order_discount.category = OrderDiscount.DEFERRED
        voucher = discount.get("voucher", None)
        if voucher:
            order_discount.voucher_id = voucher.id
            order_discount.voucher_code = voucher.code
        order_discount.save()

    def record_discount(self, discount):
        discount["offer"].record_usage(discount)
        if "voucher" in discount and discount["voucher"]:
            discount["voucher"].record_discount(discount)

    def record_voucher_usage(self, order, voucher, user):
        """
        Updates the models that care about this voucher.
        """
        voucher.record_usage(order, user)


class OrderDispatcher:
    """
    Dispatcher to send concrete order related emails.
    """

    # Event codes
    ORDER_PLACED_EVENT_CODE = "ORDER_PLACED"

    def __init__(self, logger=None, mail_connection=None):
        self.dispatcher = Dispatcher(logger=logger, mail_connection=mail_connection)

    def dispatch_order_messages(
        self, order, messages, event_code, attachments=None, **kwargs
    ):
        """
        Dispatch order-related messages to the customer.
        """
        self.dispatcher.logger.info(
            "Order #%s - sending %s messages", order.number, event_code
        )
        if order.is_anonymous:
            email = kwargs.get("email_address", order.guest_email)
            dispatched_messages = self.dispatcher.dispatch_anonymous_messages(
                email, messages, attachments
            )
        else:
            dispatched_messages = self.dispatcher.dispatch_user_messages(
                order.user, messages, attachments
            )

        try:
            event_type = CommunicationEventType.objects.get(code=event_code)
        except CommunicationEventType.DoesNotExist:
            event_type = None

        self.create_communication_event(order, event_type, dispatched_messages)

    def create_communication_event(self, order, event_type, dispatched_messages):
        """
        Create order communications event for audit.
        """
        if dispatched_messages and event_type is not None:
            CommunicationEvent._default_manager.create(
                order=order, event_type=event_type
            )

    def send_order_placed_email_for_user(self, order, extra_context, attachments=None):
        event_code = self.ORDER_PLACED_EVENT_CODE
        messages = self.dispatcher.get_messages(event_code, extra_context)
        self.dispatch_order_messages(
            order, messages, event_code, attachments=attachments
        )



# pylint: disable=F0002
import logging
from decimal import Decimal as D

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.core.signing import BadSignature, Signer
from django.db import models, transaction
from django.db.models import Sum
from django.db.models.functions import Coalesce
from django.utils import timezone
from django.utils.crypto import constant_time_compare
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.apps.order.signals import order_line_status_changed, order_status_changed
from oscar.apps.partner.exceptions import InvalidStockAdjustment
from oscar.core.compat import AUTH_USER_MODEL
from oscar.core.loading import get_model
from oscar.core.utils import get_default_currency
from oscar.models.fields import AutoSlugField

from . import exceptions

logger = logging.getLogger("oscar.order")


class AbstractOrder(models.Model):
    """
    The main order model
    """

    number = models.CharField(
        _("Order number"), max_length=128, db_index=True, unique=True
    )

    # We track the site that each order is placed within
    site = models.ForeignKey(
        "sites.Site", verbose_name=_("Site"), null=True, on_delete=models.SET_NULL
    )

    basket = models.ForeignKey(
        "basket.Basket",
        verbose_name=_("Basket"),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    # Orders can be placed without the user authenticating so we don't always
    # have a customer ID.
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        related_name="orders",
        null=True,
        blank=True,
        verbose_name=_("User"),
        on_delete=models.SET_NULL,
    )

    # Billing address is not always required (eg paying by gift card)
    billing_address = models.ForeignKey(
        "order.BillingAddress",
        null=True,
        blank=True,
        verbose_name=_("Billing Address"),
        on_delete=models.SET_NULL,
    )

    # Total price looks like it could be calculated by adding up the
    # prices of the associated lines, but in some circumstances extra
    # order-level charges are added and so we need to store it separately
    currency = models.CharField(
        _("Currency"), max_length=12, default=get_default_currency
    )
    total_incl_tax = models.DecimalField(
        _("Order total (inc. tax)"), decimal_places=2, max_digits=12
    )
    total_excl_tax = models.DecimalField(
        _("Order total (excl. tax)"), decimal_places=2, max_digits=12
    )

    # Shipping charges
    shipping_incl_tax = models.DecimalField(
        _("Shipping charge (inc. tax)"), decimal_places=2, max_digits=12, default=0
    )
    shipping_excl_tax = models.DecimalField(
        _("Shipping charge (excl. tax)"), decimal_places=2, max_digits=12, default=0
    )
    shipping_tax_code = models.CharField(
        _("Shipping VAT rate code"), max_length=64, blank=True, null=True
    )

    # Not all lines are actually shipped (such as downloads), hence shipping
    # address is not mandatory.
    shipping_address = models.ForeignKey(
        "order.ShippingAddress",
        null=True,
        blank=True,
        verbose_name=_("Shipping Address"),
        on_delete=models.SET_NULL,
    )
    shipping_method = models.CharField(_("Shipping method"), max_length=128, blank=True)

    # Identifies shipping code
    shipping_code = models.CharField(blank=True, max_length=128, default="")

    # Use this field to indicate that an order is on hold / awaiting payment
    status = models.CharField(_("Status"), max_length=100, blank=True)
    guest_email = models.EmailField(_("Guest email address"), blank=True)

    # Index added to this field for reporting
    date_placed = models.DateTimeField(db_index=True)

    # This is to prevent duplicate tracking events from being sent.
    analytics_tracked = models.BooleanField(default=False)

    #: Order status pipeline.  This should be a dict where each (key, value) #:
    #: corresponds to a status and a list of possible statuses that can follow
    #: that one.
    pipeline = getattr(settings, "OSCAR_ORDER_STATUS_PIPELINE", {})

    #: Order status cascade pipeline.  This should be a dict where each (key,
    #: value) pair corresponds to an *order* status and the corresponding
    #: *line* status that needs to be set when the order is set to the new
    #: status
    cascade = getattr(settings, "OSCAR_ORDER_STATUS_CASCADE", {})

    @classmethod
    def all_statuses(cls):
        """
        Return all possible statuses for an order
        """
        return list(cls.pipeline.keys())

    def available_statuses(self):
        """
        Return all possible statuses that this order can move to
        """
        return self.pipeline.get(self.status, ())

    def set_status(self, new_status):
        """
        Set a new status for this order.

        If the requested status is not valid, then ``InvalidOrderStatus`` is
        raised.
        """
        if new_status == self.status:
            return

        old_status = self.status

        if new_status not in self.available_statuses():
            raise exceptions.InvalidOrderStatus(
                _(
                    "'%(new_status)s' is not a valid status for order %(number)s"
                    " (current status: '%(status)s')"
                )
                % {
                    "new_status": new_status,
                    "number": self.number,
                    "status": self.status,
                }
            )
        self.status = new_status
        if new_status in self.cascade:
            new_line_status = self.cascade[new_status]
            for line in self.lines.all():
                if new_line_status in line.available_statuses():
                    line.status = new_line_status
                    line.save()
        self.save()

        # Send signal for handling status changed
        order_status_changed.send(
            sender=self,
            order=self,
            old_status=old_status,
            new_status=new_status,
        )

        self._create_order_status_change(old_status, new_status)

    set_status.alters_data = True

    def _create_order_status_change(self, old_status, new_status):
        # Not setting the status on the order as that should be handled before
        self.status_changes.create(old_status=old_status, new_status=new_status)

    @property
    def is_anonymous(self):
        # It's possible for an order to be placed by a customer who then
        # deletes their profile.  Hence, we need to check that a guest email is
        # set.
        return self.user is None and bool(self.guest_email)

    @property
    def basket_total_before_discounts_incl_tax(self):
        """
        Return basket total including tax but before discounts are applied
        """
        result = self.lines.aggregate(
            total=Sum("line_price_before_discounts_incl_tax", default=0)
        )
        return result["total"]

    @property
    def basket_total_before_discounts_excl_tax(self):
        """
        Return basket total excluding tax but before discounts are applied
        """
        result = self.lines.aggregate(
            total=Sum("line_price_before_discounts_excl_tax", default=0)
        )
        return result["total"]

    @property
    def basket_total_incl_tax(self):
        """
        Return basket total including tax
        """
        return self.total_incl_tax - self.shipping_incl_tax - self.surcharge_incl_tax

    @property
    def basket_total_excl_tax(self):
        """
        Return basket total excluding tax
        """
        return self.total_excl_tax - self.shipping_excl_tax - self.surcharge_excl_tax

    @property
    def total_before_discounts_incl_tax(self):
        return self.basket_total_before_discounts_incl_tax + self.shipping_incl_tax

    @property
    def total_before_discounts_excl_tax(self):
        return self.basket_total_before_discounts_excl_tax + self.shipping_excl_tax

    @property
    def total_discount_incl_tax(self):
        """
        The amount of discount this order received
        """
        discount = D("0.00")
        for line in self.lines.all():
            discount += line.discount_incl_tax
        return discount

    @property
    def total_discount_excl_tax(self):
        discount = D("0.00")
        for line in self.lines.all():
            discount += line.discount_excl_tax
        return discount

    @property
    def total_tax(self):
        return self.total_incl_tax - self.total_excl_tax

    @property
    def surcharge_excl_tax(self):
        return sum(charge.excl_tax for charge in self.surcharges.all())

    @property
    def surcharge_incl_tax(self):
        return sum(charge.incl_tax for charge in self.surcharges.all())

    @property
    def num_lines(self):
        return self.lines.count()

    @property
    def num_items(self):
        """
        Returns the number of items in this order.
        """
        num_items = 0
        for line in self.lines.all():
            num_items += line.quantity
        return num_items

    @property
    def shipping_tax(self):
        return self.shipping_incl_tax - self.shipping_excl_tax

    @property
    def shipping_status(self):
        """Return the last complete shipping event for this order."""

        # As safeguard against identical timestamps, also sort by the primary
        # key. It's not recommended to rely on this behaviour, but in practice
        # reasonably safe if PKs are not manually set.
        events = self.shipping_events.order_by("-date_created", "-pk").all()
        if not len(events):
            return ""

        # Collect all events by event-type
        event_map = {}
        for event in events:
            event_name = event.event_type.name
            if event_name not in event_map:
                event_map[event_name] = []
            event_map[event_name].extend(list(event.line_quantities.all()))

        # Determine last complete event
        status = _("In progress")
        for event_name, event_line_quantities in event_map.items():
            if self._is_event_complete(event_line_quantities):
                return event_name
        return status

    @property
    def has_shipping_discounts(self):
        return len(self.shipping_discounts) > 0

    @property
    def shipping_before_discounts_incl_tax(self):
        # We can construct what shipping would have been before discounts by
        # adding the discounts back onto the final shipping charge.
        total = D("0.00")
        for discount in self.shipping_discounts:
            total += discount.amount
        return self.shipping_incl_tax + total

    def _is_event_complete(self, event_quantities):
        # Form map of line to quantity
        event_map = {}
        for event_quantity in event_quantities:
            line_id = event_quantity.line_id
            event_map.setdefault(line_id, 0)
            event_map[line_id] += event_quantity.quantity

        for line in self.lines.all():
            if event_map.get(line.pk, 0) != line.quantity:
                return False
        return True

    class Meta:
        abstract = True
        app_label = "order"
        ordering = ["-date_placed"]
        verbose_name = _("Order")
        verbose_name_plural = _("Orders")

    def __str__(self):
        return "#%s" % (self.number,)

    def verification_hash(self):
        signer = Signer(salt="oscar.apps.order.Order")
        return signer.sign(self.number)

    def check_verification_hash(self, hash_to_check):
        """
        Checks the received verification hash against this order number.
        Returns False if the verification failed, True otherwise.
        """
        signer = Signer(salt="oscar.apps.order.Order")
        try:
            signed_number = signer.unsign(hash_to_check)
        except BadSignature:
            return False

        return constant_time_compare(signed_number, self.number)

    @property
    def email(self):
        if not self.user:
            return self.guest_email
        return self.user.email

    @property
    def basket_discounts(self):
        # This includes both offer- and voucher- discounts.  For orders we
        # don't need to treat them differently like we do for baskets.
        return self.discounts.filter(category=AbstractOrderDiscount.BASKET)

    @property
    def shipping_discounts(self):
        return self.discounts.filter(category=AbstractOrderDiscount.SHIPPING)

    @property
    def post_order_actions(self):
        return self.discounts.filter(category=AbstractOrderDiscount.DEFERRED)

    def set_date_placed_default(self):
        if self.date_placed is None:
            self.date_placed = now()

    def save(self, *args, **kwargs):
        # Ensure the date_placed field works as it auto_now_add was set. But
        # this gives us the ability to set the date_placed explicitly (which is
        # useful when importing orders from another system).
        self.set_date_placed_default()
        super().save(*args, **kwargs)


class AbstractOrderNote(models.Model):
    """
    A note against an order.

    This are often used for audit purposes too.  IE, whenever an admin
    makes a change to an order, we create a note to record what happened.
    """

    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="notes",
        verbose_name=_("Order"),
    )

    # These are sometimes programatically generated so don't need a
    # user everytime
    user = models.ForeignKey(
        AUTH_USER_MODEL, on_delete=models.CASCADE, null=True, verbose_name=_("User")
    )

    # We allow notes to be classified although this isn't always needed
    INFO, WARNING, ERROR, SYSTEM = "Info", "Warning", "Error", "System"
    note_type = models.CharField(_("Note Type"), max_length=128, blank=True)

    message = models.TextField(_("Message"))
    date_created = models.DateTimeField(_("Date Created"), auto_now_add=True)
    date_updated = models.DateTimeField(_("Date Updated"), auto_now=True)

    # Notes can only be edited for 5 minutes after being created
    editable_lifetime = 300

    class Meta:
        abstract = True
        app_label = "order"
        ordering = ["-date_updated"]
        verbose_name = _("Order Note")
        verbose_name_plural = _("Order Notes")

    def __str__(self):
        return "'%s' (%s)" % (self.message[0:50], self.user)

    def is_editable(self):
        if self.note_type == self.SYSTEM:
            return False
        delta = timezone.now() - self.date_updated
        return delta.seconds < self.editable_lifetime


class AbstractOrderStatusChange(models.Model):
    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="status_changes",
        verbose_name=_("Order"),
    )
    old_status = models.CharField(_("Old Status"), max_length=100, blank=True)
    new_status = models.CharField(_("New Status"), max_length=100, blank=True)
    date_created = models.DateTimeField(
        _("Date Created"), auto_now_add=True, db_index=True
    )

    class Meta:
        abstract = True
        app_label = "order"
        verbose_name = _("Order Status Change")
        verbose_name_plural = _("Order Status Changes")
        ordering = ["-date_created"]

    def __str__(self):
        return _(
            "%(order)s has changed status from %(old_status)s to %(new_status)s"
        ) % {
            "order": self.order,
            "old_status": self.old_status,
            "new_status": self.new_status,
        }


class AbstractCommunicationEvent(models.Model):
    """
    An order-level event involving a communication to the customer, such
    as an confirmation email being sent.
    """

    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="communication_events",
        verbose_name=_("Order"),
    )
    event_type = models.ForeignKey(
        "communication.CommunicationEventType",
        on_delete=models.CASCADE,
        verbose_name=_("Event Type"),
    )
    date_created = models.DateTimeField(_("Date"), auto_now_add=True, db_index=True)

    class Meta:
        abstract = True
        app_label = "order"
        verbose_name = _("Communication Event")
        verbose_name_plural = _("Communication Events")
        ordering = ["-date_created"]

    def __str__(self):
        return _("'%(type)s' event for order #%(number)s") % {
            "type": self.event_type.name,
            "number": self.order.number,
        }


# LINES


class AbstractLine(models.Model):
    """
    An order line
    """

    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="lines",
        verbose_name=_("Order"),
    )

    # PARTNER INFORMATION
    # -------------------
    # We store the partner and various detail their SKU and the title for cases
    # where the product has been deleted from the catalogue (but we still need
    # the data for reporting).  We also store the partner name in case the
    # partner gets deleted at a later date.

    partner = models.ForeignKey(
        "partner.Partner",
        related_name="order_lines",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Partner"),
    )
    partner_name = models.CharField(_("Partner name"), max_length=128, blank=True)
    partner_sku = models.CharField(_("Partner SKU"), max_length=128)

    # A line reference is the ID that a partner uses to represent this
    # particular line (it's not the same as a SKU).
    partner_line_reference = models.CharField(
        _("Partner reference"),
        max_length=128,
        blank=True,
        help_text=_(
            "This is the item number that the partner uses within their system"
        ),
    )
    partner_line_notes = models.TextField(_("Partner Notes"), blank=True)

    # We keep a link to the stockrecord used for this line which allows us to
    # update stocklevels when it ships
    stockrecord = models.ForeignKey(
        "partner.StockRecord",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Stock record"),
    )

    # PRODUCT INFORMATION
    # -------------------

    # We don't want any hard links between orders and the products table so we
    # allow this link to be NULLable.
    product = models.ForeignKey(
        "catalogue.Product",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Product"),
    )
    title = models.CharField(pgettext_lazy("Product title", "Title"), max_length=255)
    # UPC can be null because it's usually set as the product's UPC, and that
    # can be null as well
    upc = models.CharField(_("UPC"), max_length=128, blank=True, null=True)

    quantity = models.PositiveIntegerField(_("Quantity"), default=1)

    # REPORTING INFORMATION
    # ---------------------

    # Price information (these fields are actually redundant as the information
    # can be calculated from the LinePrice models
    line_price_incl_tax = models.DecimalField(
        _("Price (inc. tax)"), decimal_places=2, max_digits=12
    )
    line_price_excl_tax = models.DecimalField(
        _("Price (excl. tax)"), decimal_places=2, max_digits=12
    )

    # Price information before discounts are applied
    line_price_before_discounts_incl_tax = models.DecimalField(
        _("Price before discounts (inc. tax)"), decimal_places=2, max_digits=12
    )
    line_price_before_discounts_excl_tax = models.DecimalField(
        _("Price before discounts (excl. tax)"), decimal_places=2, max_digits=12
    )

    # Normal site price for item (without discounts)
    unit_price_incl_tax = models.DecimalField(
        _("Unit Price (inc. tax)"),
        decimal_places=2,
        max_digits=12,
        blank=True,
        null=True,
    )
    unit_price_excl_tax = models.DecimalField(
        _("Unit Price (excl. tax)"),
        decimal_places=2,
        max_digits=12,
        blank=True,
        null=True,
    )

    tax_code = models.CharField(
        _("VAT rate code"), max_length=64, blank=True, null=True
    )

    # Partners often want to assign some status to each line to help with their
    # own business processes.
    status = models.CharField(_("Status"), max_length=255, blank=True)

    num_allocated = models.PositiveIntegerField(_("Number allocated"))

    # Checks whether line allocation was cancelled or not
    allocation_cancelled = models.BooleanField(default=False)

    #: Order status pipeline.  This should be a dict where each (key, value)
    #: corresponds to a status and the possible statuses that can follow that
    #: one.
    pipeline = getattr(settings, "OSCAR_LINE_STATUS_PIPELINE", {})

    class Meta:
        abstract = True
        app_label = "order"
        # Enforce sorting in order of creation.
        ordering = ["pk"]
        verbose_name = _("Order Line")
        verbose_name_plural = _("Order Lines")

    def __str__(self):
        if self.product:
            title = self.product.title
        else:
            title = _("<missing product>")
        return _("Product '%(name)s', quantity '%(qty)s'") % {
            "name": title,
            "qty": self.quantity,
        }

    @classmethod
    def all_statuses(cls):
        """
        Return all possible statuses for an order line
        """
        return list(cls.pipeline.keys())

    def available_statuses(self):
        """
        Return all possible statuses that this order line can move to
        """
        return self.pipeline.get(self.status, ())

    def set_status(self, new_status):
        """
        Set a new status for this line

        If the requested status is not valid, then ``InvalidLineStatus`` is
        raised.
        """
        if new_status == self.status:
            return

        old_status = self.status

        if new_status not in self.available_statuses():
            raise exceptions.InvalidLineStatus(
                _(
                    "'%(new_status)s' is not a valid status (current status:"
                    " '%(status)s')"
                )
                % {"new_status": new_status, "status": self.status}
            )
        self.status = new_status
        self.save()

        # Send signal for handling status changed
        order_line_status_changed.send(
            sender=self,
            line=self,
            old_status=old_status,
            new_status=new_status,
        )

    set_status.alters_data = True

    @property
    def description(self):
        """
        Returns a description of this line including details of any
        line attributes.
        """
        desc = self.title
        ops = []
        for attribute in self.attributes.all():
            value = attribute.value
            if isinstance(value, list):
                ops.append(
                    "%s = '%s'"
                    % (attribute.option.name, (", ".join([str(v) for v in value])))
                )
            else:
                ops.append("%s = '%s'" % (attribute.option.name, value))
        if ops:
            desc = "%s (%s)" % (desc, ", ".join(ops))
        return desc

    @property
    def discount_incl_tax(self):
        return self.line_price_before_discounts_incl_tax - self.line_price_incl_tax

    @property
    def discount_excl_tax(self):
        return self.line_price_before_discounts_excl_tax - self.line_price_excl_tax

    @property
    def line_price_tax(self):
        return self.line_price_incl_tax - self.line_price_excl_tax

    @property
    def unit_price_tax(self):
        return self.unit_price_incl_tax - self.unit_price_excl_tax

    # Shipping status helpers

    @property
    def shipping_status(self):
        """
        Returns a string summary of the shipping status of this line
        """
        status_map = self.shipping_event_breakdown
        if not status_map:
            return ""

        events = []
        last_complete_event_name = None
        for event_dict in reversed(list(status_map.values())):
            if event_dict["quantity"] == self.quantity:
                events.append(event_dict["name"])
                last_complete_event_name = event_dict["name"]
            else:
                events.append(
                    "%s (%d/%d items)"
                    % (event_dict["name"], event_dict["quantity"], self.quantity)
                )

        if last_complete_event_name == list(status_map.values())[0]["name"]:
            return last_complete_event_name

        return ", ".join(events)

    def is_shipping_event_permitted(self, event_type, quantity):
        """
        Test whether a shipping event with the given quantity is permitted

        This method should normally be overridden to ensure that the
        prerequisite shipping events have been passed for this line.
        """
        # Note, this calculation is simplistic - normally, you will also need
        # to check if previous shipping events have occurred.  Eg, you can't
        # return lines until they have been shipped.
        current_qty = self.shipping_event_quantity(event_type)
        return (current_qty + quantity) <= self.quantity

    def shipping_event_quantity(self, event_type):
        """
        Return the quantity of this line that has been involved in a shipping
        event of the passed type.
        """
        result = self.shipping_event_quantities.filter(
            event__event_type=event_type
        ).aggregate(Sum("quantity"))
        if result["quantity__sum"] is None:
            return 0
        else:
            return result["quantity__sum"]

    def has_shipping_event_occurred(self, event_type, quantity=None):
        """
        Test whether this line has passed a given shipping event
        """
        if not quantity:
            quantity = self.quantity
        return self.shipping_event_quantity(event_type) == quantity

    def get_event_quantity(self, event):
        """
        Fetches the ShippingEventQuantity instance for this line

        Exists as a separate method so it can be overridden to avoid
        the DB query that's caused by get().
        """
        return event.line_quantities.get(line=self)

    @property
    def shipping_event_breakdown(self):
        """
        Returns a dict of shipping events that this line has been through
        """
        status_map = {}
        for event in self.shipping_events.all():
            event_type = event.event_type
            event_name = event_type.name
            event_quantity = self.get_event_quantity(event).quantity
            if event_name in status_map:
                status_map[event_name]["quantity"] += event_quantity
            else:
                status_map[event_name] = {
                    "event_type": event_type,
                    "name": event_name,
                    "quantity": event_quantity,
                }
        return status_map

    # Payment event helpers

    def is_payment_event_permitted(self, event_type, quantity):
        """
        Test whether a payment event with the given quantity is permitted.

        Allow each payment event type to occur only once per quantity.
        """
        current_qty = self.payment_event_quantity(event_type)
        return (current_qty + quantity) <= self.quantity

    def payment_event_quantity(self, event_type):
        """
        Return the quantity of this line that has been involved in a payment
        event of the passed type.
        """
        result = self.payment_event_quantities.filter(
            event__event_type=event_type
        ).aggregate(Sum("quantity"))
        if result["quantity__sum"] is None:
            return 0
        else:
            return result["quantity__sum"]

    @property
    def is_product_deleted(self):
        return self.product is None

    def is_available_to_reorder(self, basket, strategy):
        """
        Test if this line can be re-ordered using the passed strategy and
        basket
        """
        if not self.product:
            return False, (
                _("'%(title)s' is no longer available") % {"title": self.title}
            )

        try:
            basket_line = basket.lines.get(product=self.product)
        except basket.lines.model.DoesNotExist:
            desired_qty = self.quantity
        else:
            desired_qty = basket_line.quantity + self.quantity

        result = strategy.fetch_for_product(self.product)
        is_available, reason = result.availability.is_purchase_permitted(
            quantity=desired_qty
        )
        if not is_available:
            return False, reason
        return True, None

    @property
    def can_track_allocations(self):
        return self.stockrecord.can_track_allocations

    def is_allocation_consumption_possible(self, quantity):
        if self.allocation_cancelled:
            return False

        return quantity <= self.num_allocated

    def consume_allocation(self, quantity):
        if not self.can_track_allocations:
            return
        if not self.is_allocation_consumption_possible(quantity):
            raise InvalidStockAdjustment(_("Invalid stock consumption request"))

        with transaction.atomic():
            self.__class__.objects.filter(pk=self.pk).update(
                num_allocated=(Coalesce(models.F("num_allocated"), 0) - quantity),
            )
            if self.stockrecord:
                self.stockrecord.consume_allocation(quantity)
        self.refresh_from_db(fields=["num_allocated"])

    consume_allocation.alters_data = True

    def cancel_allocation(self, quantity):
        if not self.can_track_allocations:
            return
        if not self.is_allocation_consumption_possible(quantity):
            return

        with transaction.atomic():
            locked_self = (
                self.__class__.objects.filter(pk=self.pk).select_for_update().get()
            )
            if locked_self.num_allocated == quantity:
                locked_self.num_allocated = 0
                locked_self.allocation_cancelled = True
            else:
                locked_self.num_allocated -= quantity

            locked_self.save()
            if locked_self.stockrecord:
                locked_self.stockrecord.cancel_allocation(quantity)

        self.refresh_from_db(fields=["num_allocated", "allocation_cancelled"])

    cancel_allocation.alters_data = True


class AbstractLineAttribute(models.Model):
    """
    An attribute of a line
    """

    line = models.ForeignKey(
        "order.Line",
        on_delete=models.CASCADE,
        related_name="attributes",
        verbose_name=_("Line"),
    )
    option = models.ForeignKey(
        "catalogue.Option",
        null=True,
        on_delete=models.SET_NULL,
        related_name="line_attributes",
        verbose_name=_("Option"),
    )
    type = models.CharField(_("Type"), max_length=128)
    value = models.JSONField(_("Value"), encoder=DjangoJSONEncoder)

    class Meta:
        abstract = True
        app_label = "order"
        verbose_name = _("Line Attribute")
        verbose_name_plural = _("Line Attributes")

    def __str__(self):
        return "%s = %s" % (self.type, self.value)


class AbstractLinePrice(models.Model):
    """
    For tracking the prices paid for each unit within a line.

    This is necessary as offers can lead to units within a line
    having different prices.  For example, one product may be sold at
    50% off as it's part of an offer while the remainder are full price.
    """

    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="line_prices",
        verbose_name=_("Order"),
    )
    line = models.ForeignKey(
        "order.Line",
        on_delete=models.CASCADE,
        related_name="prices",
        verbose_name=_("Line"),
    )
    quantity = models.PositiveIntegerField(_("Quantity"), default=1)
    price_incl_tax = models.DecimalField(
        _("Price (inc. tax)"), decimal_places=2, max_digits=12
    )
    price_excl_tax = models.DecimalField(
        _("Price (excl. tax)"), decimal_places=2, max_digits=12
    )
    shipping_incl_tax = models.DecimalField(
        _("Shiping (inc. tax)"), decimal_places=2, max_digits=12, default=0
    )
    shipping_excl_tax = models.DecimalField(
        _("Shipping (excl. tax)"), decimal_places=2, max_digits=12, default=0
    )
    tax_code = models.CharField(
        _("VAT rate code"), max_length=64, blank=True, null=True
    )

    class Meta:
        abstract = True
        app_label = "order"
        ordering = ("id",)
        verbose_name = _("Line Price")
        verbose_name_plural = _("Line Prices")

    def __str__(self):
        return _("Line '%(number)s' (quantity %(qty)d) price %(price)s") % {
            "number": self.line,
            "qty": self.quantity,
            "price": self.price_incl_tax,
        }


# PAYMENT EVENTS


class AbstractPaymentEventType(models.Model):
    """
    Payment event types are things like 'Paid', 'Failed', 'Refunded'.

    These are effectively the transaction types.
    """

    name = models.CharField(_("Name"), max_length=128, unique=True)
    code = AutoSlugField(_("Code"), max_length=128, unique=True, populate_from="name")

    class Meta:
        abstract = True
        app_label = "order"
        verbose_name = _("Payment Event Type")
        verbose_name_plural = _("Payment Event Types")
        ordering = ("name",)

    def __str__(self):
        return self.name


class AbstractPaymentEvent(models.Model):
    """
    A payment event for an order

    For example:

    * All lines have been paid for
    * 2 lines have been refunded
    """

    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="payment_events",
        verbose_name=_("Order"),
    )
    amount = models.DecimalField(_("Amount"), decimal_places=2, max_digits=12)
    # The reference should refer to the transaction ID of the payment gateway
    # that was used for this event.
    reference = models.CharField(_("Reference"), max_length=128, blank=True)
    lines = models.ManyToManyField(
        "order.Line", through="PaymentEventQuantity", verbose_name=_("Lines")
    )
    event_type = models.ForeignKey(
        "order.PaymentEventType", on_delete=models.CASCADE, verbose_name=_("Event Type")
    )
    # Allow payment events to be linked to shipping events.  Often a shipping
    # event will trigger a payment event and so we can use this FK to capture
    # the relationship.
    shipping_event = models.ForeignKey(
        "order.ShippingEvent",
        null=True,
        on_delete=models.CASCADE,
        related_name="payment_events",
    )
    date_created = models.DateTimeField(
        _("Date created"), auto_now_add=True, db_index=True
    )

    class Meta:
        abstract = True
        app_label = "order"
        verbose_name = _("Payment Event")
        verbose_name_plural = _("Payment Events")
        ordering = ["-date_created"]

    def __str__(self):
        return _("Payment event for order %s") % self.order

    def num_affected_lines(self):
        return self.lines.all().count()


class PaymentEventQuantity(models.Model):
    """
    A "through" model linking lines to payment events
    """

    event = models.ForeignKey(
        "order.PaymentEvent",
        on_delete=models.CASCADE,
        related_name="line_quantities",
        verbose_name=_("Event"),
    )
    line = models.ForeignKey(
        "order.Line",
        on_delete=models.CASCADE,
        related_name="payment_event_quantities",
        verbose_name=_("Line"),
    )
    quantity = models.PositiveIntegerField(_("Quantity"))

    class Meta:
        app_label = "order"
        verbose_name = _("Payment Event Quantity")
        verbose_name_plural = _("Payment Event Quantities")
        unique_together = ("event", "line")


# SHIPPING EVENTS


class AbstractShippingEvent(models.Model):
    """
    An event is something which happens to a group of lines such as
    1 item being dispatched.
    """

    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="shipping_events",
        verbose_name=_("Order"),
    )
    lines = models.ManyToManyField(
        "order.Line",
        related_name="shipping_events",
        through="ShippingEventQuantity",
        verbose_name=_("Lines"),
    )
    event_type = models.ForeignKey(
        "order.ShippingEventType",
        on_delete=models.CASCADE,
        verbose_name=_("Event Type"),
    )
    notes = models.TextField(
        _("Event notes"),
        blank=True,
        help_text=_("This could be the dispatch reference, or a tracking number"),
    )
    date_created = models.DateTimeField(
        _("Date Created"), auto_now_add=True, db_index=True
    )

    class Meta:
        abstract = True
        app_label = "order"
        verbose_name = _("Shipping Event")
        verbose_name_plural = _("Shipping Events")
        ordering = ["-date_created"]

    def __str__(self):
        return _("Order #%(number)s, type %(type)s") % {
            "number": self.order.number,
            "type": self.event_type,
        }

    def num_affected_lines(self):
        return self.lines.count()


class ShippingEventQuantity(models.Model):
    """
    A "through" model linking lines to shipping events.

    This exists to track the quantity of a line that is involved in a
    particular shipping event.
    """

    event = models.ForeignKey(
        "order.ShippingEvent",
        on_delete=models.CASCADE,
        related_name="line_quantities",
        verbose_name=_("Event"),
    )
    line = models.ForeignKey(
        "order.Line",
        on_delete=models.CASCADE,
        related_name="shipping_event_quantities",
        verbose_name=_("Line"),
    )
    quantity = models.PositiveIntegerField(_("Quantity"))

    class Meta:
        app_label = "order"
        verbose_name = _("Shipping Event Quantity")
        verbose_name_plural = _("Shipping Event Quantities")
        unique_together = ("event", "line")

    def save(self, *args, **kwargs):
        # Default quantity to full quantity of line
        if not self.quantity:
            self.quantity = self.line.quantity
        # Ensure we don't violate quantities constraint
        if not self.line.is_shipping_event_permitted(
            self.event.event_type, self.quantity
        ):
            raise exceptions.InvalidShippingEvent
        super().save(*args, **kwargs)

    def __str__(self):
        return _("%(product)s - quantity %(qty)d") % {
            "product": self.line.product,
            "qty": self.quantity,
        }


class AbstractShippingEventType(models.Model):
    """
    A type of shipping/fulfilment event

    E.g.: 'Shipped', 'Cancelled', 'Returned'
    """

    # Name is the friendly description of an event
    name = models.CharField(_("Name"), max_length=255, unique=True)
    # Code is used in forms
    code = AutoSlugField(_("Code"), max_length=128, unique=True, populate_from="name")

    class Meta:
        abstract = True
        app_label = "order"
        verbose_name = _("Shipping Event Type")
        verbose_name_plural = _("Shipping Event Types")
        ordering = ("name",)

    def __str__(self):
        return self.name


# DISCOUNTS


class AbstractOrderDiscount(models.Model):
    """
    A discount against an order.

    Normally only used for display purposes so an order can be listed with
    discounts displayed separately even though in reality, the discounts are
    applied at the line level.

    This has evolved to be a slightly misleading class name as this really
    track benefit applications which aren't necessarily discounts.
    """

    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="discounts",
        verbose_name=_("Order"),
    )

    # We need to distinguish between basket discounts, shipping discounts and
    # 'deferred' discounts.
    BASKET, SHIPPING, DEFERRED = "Basket", "Shipping", "Deferred"
    CATEGORY_CHOICES = (
        (BASKET, _(BASKET)),
        (SHIPPING, _(SHIPPING)),
        (DEFERRED, _(DEFERRED)),
    )
    category = models.CharField(
        _("Discount category"), default=BASKET, max_length=64, choices=CATEGORY_CHOICES
    )

    offer_id = models.PositiveIntegerField(_("Offer ID"), blank=True, null=True)
    offer_name = models.CharField(
        _("Offer name"), max_length=128, db_index=True, blank=True
    )
    voucher_id = models.PositiveIntegerField(_("Voucher ID"), blank=True, null=True)
    voucher_code = models.CharField(
        _("Code"), max_length=128, db_index=True, blank=True
    )
    frequency = models.PositiveIntegerField(_("Frequency"), null=True)
    amount = models.DecimalField(
        _("Amount"), decimal_places=2, max_digits=12, default=0
    )

    # Post-order offer applications can return a message to indicate what
    # action was taken after the order was placed.
    message = models.TextField(blank=True)

    @property
    def is_basket_discount(self):
        return self.category == self.BASKET

    @property
    def is_shipping_discount(self):
        return self.category == self.SHIPPING

    @property
    def is_post_order_action(self):
        return self.category == self.DEFERRED

    class Meta:
        abstract = True
        app_label = "order"
        ordering = ["pk"]
        verbose_name = _("Order Discount")
        verbose_name_plural = _("Order Discounts")

    def save(self, *args, **kwargs):
        if self.offer_id and not self.offer_name:
            offer = self.offer
            if offer:
                self.offer_name = offer.name

        if self.voucher_id and not self.voucher_code:
            voucher = self.voucher
            if voucher:
                self.voucher_code = voucher.code

        super().save(**kwargs)

    def __str__(self):
        return _("Discount of %(amount)r from order %(order)s") % {
            "amount": self.amount,
            "order": self.order,
        }

    @property
    def offer(self):
        Offer = get_model("offer", "ConditionalOffer")
        try:
            return Offer.objects.get(id=self.offer_id)
        except Offer.DoesNotExist:
            return None

    @property
    def voucher(self):
        Voucher = get_model("voucher", "Voucher")
        try:
            return Voucher.objects.get(id=self.voucher_id)
        except Voucher.DoesNotExist:
            return None

    def description(self):
        if self.voucher_code:
            return self.voucher_code
        return self.offer_name or ""


class AbstractOrderLineDiscount(models.Model):
    line = models.ForeignKey(
        "order.Line",
        on_delete=models.CASCADE,
        related_name="discounts",
        verbose_name=_("Line"),
    )
    order_discount = models.ForeignKey(
        "order.OrderDiscount",
        on_delete=models.CASCADE,
        related_name="discount_lines",
        verbose_name=_("Order discount"),
    )

    is_incl_tax = models.BooleanField()
    amount = models.DecimalField(
        _("Line discount (excl. tax)"), decimal_places=2, max_digits=12, default=0
    )

    class Meta:
        abstract = True
        app_label = "order"
        ordering = ["pk"]
        verbose_name = _("Order line discount")
        verbose_name_plural = _("Order line discounts")


class AbstractSurcharge(models.Model):
    order = models.ForeignKey(
        "order.Order",
        on_delete=models.CASCADE,
        related_name="surcharges",
        verbose_name=_("Order"),
    )

    name = models.CharField(_("Name"), max_length=128)

    code = models.CharField(_("Surcharge code"), max_length=128)

    incl_tax = models.DecimalField(
        _("Surcharge (inc. tax)"), decimal_places=2, max_digits=12, default=0
    )
    excl_tax = models.DecimalField(
        _("Surcharge (excl. tax)"), decimal_places=2, max_digits=12, default=0
    )
    tax_code = models.CharField(
        _("VAT rate code"), max_length=64, blank=True, null=True
    )

    @property
    def tax(self):
        return self.incl_tax - self.excl_tax

    class Meta:
        abstract = True
        app_label = "order"
        ordering = ["pk"]
        verbose_name = _("Surcharge")
        verbose_name_plural = _("Surcharges")



import logging

from django.db.models import Max
from django.template import loader

from oscar.core.loading import get_class, get_model

ProductAlert = get_model("customer", "ProductAlert")
Product = get_model("catalogue", "Product")
Dispatcher = get_class("communication.utils", "Dispatcher")
Selector = get_class("partner.strategy", "Selector")

alerts_logger = logging.getLogger("oscar.alerts")


class AlertsDispatcher:
    """
    Dispatcher to send concrete product alerts related emails
    and notifications.
    """

    # Event codes
    PRODUCT_ALERT_EVENT_CODE = "PRODUCT_ALERT"
    PRODUCT_ALERT_CONFIRMATION_EVENT_CODE = "PRODUCT_ALERT_CONFIRMATION"

    def __init__(self, logger=None, mail_connection=None):
        self.dispatcher = Dispatcher(
            logger=logger or alerts_logger,
            mail_connection=mail_connection,
        )

    def get_queryset(self):
        return (
            Product.objects.browsable()
            .filter(productalert__status=ProductAlert.ACTIVE)
            .distinct()
        )

    def send_alerts(self):
        """
        Check all products with active product alerts for
        availability and send out email alerts when a product is
        available to buy.
        """
        products = self.get_queryset()
        self.dispatcher.logger.info(
            "Found %d products with active alerts", products.count()
        )
        for product in products:
            self.send_product_alert_email_for_user(product)

    def get_extra_context(self, alert):
        return {
            "alert": alert,
        }

    def get_extra_context_for_alert(self, alert, hurry_mode):
        context = self.get_extra_context(alert)
        context["hurry"] = hurry_mode
        return context

    def send_product_alert_email_for_user(self, product):
        """
        Check for notifications for this product and send email to users
        if the product is back in stock. Add a little 'hurry' note if the
        amount of in-stock items is less then the number of notifications.
        """
        stockrecords = product.stockrecords.all()
        num_stockrecords = len(stockrecords)
        if not num_stockrecords:
            return

        self.dispatcher.logger.info("Sending alerts for '%s'", product)
        alerts = ProductAlert.objects.filter(
            product_id__in=(product.id, product.parent_id),
            status=ProductAlert.ACTIVE,
        )

        # Determine 'hurry mode'
        if num_stockrecords == 1:
            num_in_stock = stockrecords[0].num_in_stock
        else:
            result = stockrecords.aggregate(max_in_stock=Max("num_in_stock"))
            num_in_stock = result["max_in_stock"]

        # 'hurry_mode' is false if 'num_in_stock' is None
        hurry_mode = num_in_stock is not None and alerts.count() > num_in_stock

        messages_to_send = []
        user_messages_to_send = []
        num_notifications = 0
        selector = Selector()
        for alert in alerts:
            # Check if the product is available to this user
            strategy = selector.strategy(user=alert.user)
            data = strategy.fetch_for_product(product)
            if not data.availability.is_available_to_buy:
                continue

            extra_context = self.get_extra_context_for_alert(alert, hurry_mode)
            if alert.user:
                # Send a site notification
                num_notifications += 1
                self.notify_user_about_product_alert(alert.user, extra_context)

            messages = self.dispatcher.get_messages(
                self.PRODUCT_ALERT_EVENT_CODE, extra_context
            )

            if messages and messages["body"]:
                if alert.user:
                    user_messages_to_send.append((alert.user, messages))
                else:
                    messages_to_send.append((alert.get_email_address(), messages))
            alert.close()

        if messages_to_send or user_messages_to_send:
            for message in messages_to_send:
                self.dispatcher.dispatch_direct_messages(*message)
            for message in user_messages_to_send:
                self.dispatcher.dispatch_user_messages(*message)

        self.dispatcher.logger.info(
            "Sent %d notifications and %d messages",
            num_notifications,
            len(messages_to_send) + len(user_messages_to_send),
        )

    def send_product_alert_confirmation_email_for_user(self, alert, extra_context=None):
        """
        Send an alert confirmation email.
        """
        if extra_context is None:
            extra_context = self.get_extra_context(alert)
        messages = self.dispatcher.get_messages(
            self.PRODUCT_ALERT_CONFIRMATION_EVENT_CODE, extra_context
        )
        self.dispatcher.dispatch_direct_messages(alert.email, messages)

    def notify_user_about_product_alert(self, user, context):
        subj_tpl = loader.get_template("oscar/customer/alerts/message_subject.html")
        message_tpl = loader.get_template("oscar/customer/alerts/message.html")
        self.dispatcher.notify_user(
            user,
            subj_tpl.render(context).strip(),
            body=message_tpl.render(context).strip(),
        )



class TaxNotKnown(Exception):
    """
    Exception for when a tax-inclusive price is requested but we don't know
    what the tax applicable is (yet).
    """


class Price(object):
    """
    Simple price class that encapsulates a price and its tax information

    Attributes:
        incl_tax (Decimal): Price including taxes
        excl_tax (Decimal): Price excluding taxes
        tax (Decimal): Tax amount
        is_tax_known (bool): Whether tax is known
        currency (str): 3 character currency code
    """

    def __init__(self, currency, excl_tax, incl_tax=None, tax=None, tax_code=None):
        self.currency = currency
        self.excl_tax = excl_tax
        self.tax_code = tax_code
        if incl_tax is not None:
            self.incl_tax = incl_tax
            self.is_tax_known = True
        elif tax is not None:
            self.incl_tax = excl_tax + tax
            self.is_tax_known = True
        else:
            self.incl_tax = None
            self.is_tax_known = False

    def _get_tax(self):
        return self.incl_tax - self.excl_tax

    def _set_tax(self, value):
        self.incl_tax = self.excl_tax + value
        self.is_tax_known = True

    tax = property(_get_tax, _set_tax)

    def __repr__(self):
        if self.is_tax_known:
            return "%s(currency=%r, excl_tax=%r, incl_tax=%r, tax=%r)" % (
                self.__class__.__name__,
                self.currency,
                self.excl_tax,
                self.incl_tax,
                self.tax,
            )
        return "%s(currency=%r, excl_tax=%r)" % (
            self.__class__.__name__,
            self.currency,
            self.excl_tax,
        )

    def __eq__(self, other):
        """
        Two price objects are equal if currency, price.excl_tax and tax match.
        """
        return (
            self.currency == other.currency
            and self.excl_tax == other.excl_tax
            and self.incl_tax == other.incl_tax
        )

    def __add__(self, other):
        if self.currency != other.currency:
            raise ValueError("Cannot add prices with different currencies.")

        return Price(
            currency=self.currency,
            incl_tax=self.incl_tax + other.incl_tax,
            excl_tax=self.excl_tax + other.excl_tax,
            tax_code=self.tax_code,
        )

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)



from django.conf import settings
from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.core.paginator import Paginator
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView

from oscar.core.loading import get_model

WishList = get_model("wishlists", "WishList")


class WishListView(DetailView):
    template_name = "oscar/wishlists/wishlist_detail.html"
    model = WishList
    context_object_name = "wishlist"

    def dispatch(self, request, *args, **kwargs):
        wishlist = self.get_object()

        if wishlist.is_allowed_to_see(request.user):
            return super().dispatch(request, *args, **kwargs)
        elif (
            wishlist.visibility == WishList.PRIVATE
            or request.user.is_authenticated
            and not wishlist.is_allowed_to_see(request.user)
        ):
            raise PermissionDenied
        else:
            messages.info(request, _("You must be logged in to view the wish list"))
            redirect_url = "%s?next=%s" % (settings.LOGIN_URL, request.path)
            return redirect(redirect_url)

    def get_object(self, queryset=None):
        return get_object_or_404(WishList, key=self.kwargs.get("key"))

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        wishlist_lines = self.object.lines.filter(product__isnull=False)
        paginator = Paginator(wishlist_lines, settings.OSCAR_PRODUCTS_PER_PAGE)
        page_number = self.request.GET.get("page")
        page_obj = paginator.get_page(page_number)

        ctx["wishlist_title"] = str(self.object)
        ctx["page_obj"] = page_obj
        ctx["paginator"] = paginator

        return ctx



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-13 08:44
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0004_auto_20160111_1108'),
    ]

    operations = [
        migrations.AlterField(
            model_name='order',
            name='guest_email',
            field=models.EmailField(blank=True, max_length=254, verbose_name='Guest email address'),
        ),
    ]



"""Permissions used for different dashboard views."""

from django.contrib.auth import get_user_model

User = get_user_model()


class DashboardPermission:
    """Permissions used for different dashboard views."""

    permissions = {
        # Catalogue
        "product": [
            "is_staff",
            "catalogue.add_product",
            "catalogue.change_product",
            "catalogue.view_product",
            "catalogue.delete_product",
        ],
        "category": [
            "is_staff",
            "catalogue.add_category",
            "catalogue.change_category",
            "catalogue.view_category",
            "catalogue.delete_category",
        ],
        "product_class": [
            "is_staff",
            "catalogue.add_productclass",
            "catalogue.change_productclass",
            "catalogue.view_productclass",
            "catalogue.delete_productclass",
        ],
        "attribute_option_group": [
            "is_staff",
            "catalogue.add_attributeoptiongroup",
            "catalogue.change_attributeoptiongroup",
            "catalogue.view_attributeoptiongroup",
            "catalogue.delete_attributeoptiongroup",
        ],
        "option": [
            "is_staff",
            "catalogue.add_option",
            "catalogue.change_option",
            "catalogue.view_option",
            "catalogue.delete_option",
        ],
        # Communications
        "communication_event_type": [
            "is_staff",
            "communication.add_communicationeventtype",
            "communication.change_communicationeventtype",
            "communication.view_communicationeventtype",
            "communication.delete_communicationeventtype",
        ],
        # Partners
        "stockalert": [
            "is_staff",
            "partner.add_stockalert",
            "partner.change_stockalert",
            "partner.view_stockalert",
            "partner.delete_stockalert",
        ],
        "partner": [
            "is_staff",
            "partner.add_partner",
            "partner.change_partner",
            "partner.view_partner",
            "partner.delete_partner",
        ],
        # Offers
        "offer": [
            "is_staff",
            "offer.add_conditionaloffer",
            "offer.change_conditionaloffer",
            "offer.view_conditionaloffer",
            "offer.delete_conditionaloffer",
        ],
        # Orders
        "order": [
            "is_staff",
            "order.add_order",
            "order.change_order",
            "order.view_order",
            "order.delete_order",
        ],
        # Flat Pages
        "flat_page": [
            "is_staff",
            "flatpages.add_flatpage",
            "flatpages.change_flatpage",
            "flatpages.view_flatpage",
            "flatpages.delete_flatpage",
        ],
        # Ranges
        "range": [
            "is_staff",
            "offer.add_range",
            "offer.change_range",
            "offer.view_range",
            "offer.delete_range",
        ],
        # Analytics
        "user_record": [
            "is_staff",
            "analytics.add_userrecord",
            "analytics.change_userrecord",
            "analytics.view_userrecord",
            "analytics.delete_userrecord",
        ],
        # Reviews
        "product_review": [
            "is_staff",
            "reviews.add_productreview",
            "reviews.change_productreview",
            "reviews.view_productreview",
            "reviews.delete_productreview",
        ],
        # Shipping
        "shipping_method": [
            "is_staff",
            "shipping.add_weightbased",
            "shipping.change_weightbased",
            "shipping.delete_weightbased",
            "shipping.view_weightbased",
        ],
        # Users
        "user": [
            "is_staff",
            f"{User._meta.app_label}.add_user",
            f"{User._meta.app_label}.change_user",
            f"{User._meta.app_label}.view_user",
            f"{User._meta.app_label}.delete_user",
        ],
        # Vouchers
        "voucher": [
            "is_staff",
            "voucher.add_voucher",
            "voucher.change_voucher",
            "voucher.view_voucher",
            "voucher.delete_voucher",
        ],
    }
    # Staff
    staff = ["is_staff"]
    # Partner Access
    partner_dashboard_access = ["partner.dashboard_access"]

    @classmethod
    def get(cls, *args):
        """
        DashboardPermission.get("product", "category", "stockalert")

        :return: list of permission codes
        """
        permissions = set()
        for arg in args:
            permissions.update(cls.permissions[arg])
        permissions.remove("is_staff")
        return list(permissions)

    @classmethod
    def get_all_permissions(cls):
        """
        Retrieve set of unique permissions defined in this class.
        """
        return set(
            [
                permission
                for permissions in cls.permissions.values()
                for permission in permissions
            ]
        )

    @classmethod
    def has_dashboard_perms(cls, user):
        """
        Check if user has any of the dashboard permissions.
        """
        return len(cls.get_all_permissions() & user.get_all_permissions()) > 0
